@@ -97,12 +97,9 @@ itrunc(struct context *ctx, const struct type *type, uint64_t val)
 		return (unsigned int)val;
 	case STORAGE_ARRAY:
 	case STORAGE_ICONST:
-	case STORAGE_POINTER:
 	case STORAGE_SIZE:
 	case STORAGE_UINTPTR:
 		return val;
-	case STORAGE_BOOL:
-		return (bool)val;
 	case STORAGE_NULL:
 		return (uintptr_t)NULL;
 	case STORAGE_ALIAS:
@@ -111,12 +108,16 @@ itrunc(struct context *ctx, const struct type *type, uint64_t val)
 		return itrunc(ctx, type->alias.type, val);
 	case STORAGE_ERROR:
 		return val;
+	case STORAGE_BOOL:
+	case STORAGE_DONE:
 	case STORAGE_F32:
 	case STORAGE_F64:
 	case STORAGE_FCONST:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
+	case STORAGE_POINTER:
 	case STORAGE_SLICE:
 	case STORAGE_STRING:
 	case STORAGE_STRUCT:
@@ -125,7 +126,6 @@ itrunc(struct context *ctx, const struct type *type, uint64_t val)
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
-	case STORAGE_DONE:
 		assert(0);
 	}
 	assert(0);
@@ -325,12 +325,13 @@ eval_binarithm(struct context *ctx,
 		neg = true;
 		/* fallthrough */
 	case BIN_LEQUAL:
-		if (type_is_float(ctx, lvalue.result)) {
+		if (type_dealias(ctx, lvalue.result)->storage == STORAGE_POINTER) {
+			return false;
+		} else if (type_is_float(ctx, lvalue.result)) {
 			bval = ftrunc(ctx, lvalue.result, flval) == ftrunc(ctx, rvalue.result, frval);
 		} else if (type_is_signed(ctx, lvalue.result)) {
 			bval = itrunc(ctx, lvalue.result, ilval) == itrunc(ctx, rvalue.result, irval);
-		} else if (type_is_integer(ctx, lvalue.result)
-				|| type_dealias(ctx, lvalue.result)->storage == STORAGE_POINTER) {
+		} else if (type_is_integer(ctx, lvalue.result)) {
 			bval = itrunc(ctx, lvalue.result, ulval) == itrunc(ctx, rvalue.result, urval);
 		} else if (type_dealias(ctx, lvalue.result)->storage == STORAGE_BOOL) {
 			bval = lvalue.literal.bval == rvalue.literal.bval;
@@ -387,9 +388,10 @@ eval_binarithm(struct context *ctx,
 	} else if (type_dealias(ctx, in->result)->storage == STORAGE_BOOL
 			|| type_dealias(ctx, in->result)->storage == STORAGE_STRING) {
 		out->literal.bval = bval;
+	} else if (type_dealias(ctx, in->result)->storage == STORAGE_POINTER) {
+		return false;
 	} else {
-		assert(type_is_integer(ctx, in->result)
-			|| type_dealias(ctx, in->result)->storage == STORAGE_POINTER);
+		assert(type_is_integer(ctx, in->result));
 		out->literal.uval = itrunc(ctx, in->result, uval);
 	}
 	return true;
@@ -464,18 +466,26 @@ eval_literal(struct context *ctx,
 		}
 		break;
 	case STORAGE_BOOL:
+	case STORAGE_DONE:
 	case STORAGE_ERROR:
-	case STORAGE_F32:
 	case STORAGE_F64:
 	case STORAGE_FCONST:
+	case STORAGE_NOMEM:
+	case STORAGE_NULL:
+	case STORAGE_POINTER:
+	case STORAGE_SLICE:
+	case STORAGE_VOID:
+		out->literal = in->literal;
+		break;
+	case STORAGE_F32:
+		out->literal.fval = (float)in->literal.fval;
+		break;
 	case STORAGE_I16:
 	case STORAGE_I32:
 	case STORAGE_I64:
 	case STORAGE_I8:
 	case STORAGE_ICONST:
 	case STORAGE_INT:
-	case STORAGE_NULL:
-	case STORAGE_POINTER:
 	case STORAGE_RCONST:
 	case STORAGE_RUNE:
 	case STORAGE_SIZE:
@@ -485,10 +495,7 @@ eval_literal(struct context *ctx,
 	case STORAGE_U8:
 	case STORAGE_UINT:
 	case STORAGE_UINTPTR:
-	case STORAGE_VOID:
-	case STORAGE_DONE:
-	case STORAGE_SLICE:
-		out->literal = in->literal;
+		out->literal.uval = itrunc(ctx, in->result, in->literal.uval);
 		break;
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
@@ -614,7 +621,9 @@ eval_cast(struct context *ctx,
 	case STORAGE_SIZE:
 	case STORAGE_RCONST:
 	case STORAGE_RUNE:
-		if (type_is_float(ctx, val.result)) {
+		if (type_dealias(ctx, from)->storage == STORAGE_POINTER) {
+			return false;
+		} else if (type_is_float(ctx, val.result)) {
 			out->literal.ival =
 				itrunc(ctx, to, (int64_t)val.literal.fval);
 		} else if (type_is_signed(ctx, val.result)) {
@@ -676,9 +685,10 @@ eval_cast(struct context *ctx,
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
 		assert(0); // Invariant
-	case STORAGE_ERROR:
-	case STORAGE_VOID:
 	case STORAGE_DONE:
+	case STORAGE_ERROR:
+	case STORAGE_NOMEM:
+	case STORAGE_VOID:
 		return true;
 	}
 
@@ -691,7 +701,7 @@ eval_len(struct context *ctx,
 	struct expression *restrict out)
 {
 	assert(in->type == EXPR_LEN);
-	const struct type *expr_type = type_dereference(ctx, in->len.value->result);
+	const struct type *expr_type = type_dereference(ctx, in->len.value->result, false);
 	assert(expr_type != NULL);
 	expr_type = type_dealias(ctx, expr_type);
 
@@ -728,7 +738,8 @@ static bool
 literal_default(struct context *ctx, struct expression *v)
 {
 	struct expression b = {0};
-	switch (type_dealias(ctx, v->result)->storage) {
+	const struct type *t = type_dealias(ctx, v->result);
+	switch (t->storage) {
 	case STORAGE_ERROR:
 	case STORAGE_POINTER:
 	case STORAGE_I16:
@@ -767,12 +778,21 @@ literal_default(struct context *ctx, struct expression *v)
 		v->literal.string.len = 0;
 		break;
 	case STORAGE_ARRAY:
-		v->literal.array = xcalloc(1, sizeof(struct array_literal));
-		v->literal.array->value = xcalloc(1, sizeof(struct expression));
-		v->literal.array->value->type = EXPR_LITERAL;
-		v->literal.array->value->result =
-			type_dealias(ctx, v->result)->array.members;
-		return literal_default(ctx, v->literal.array->value);
+		assert(!t->array.expandable); // Invariant
+		if (t->array.length == SIZE_UNDEFINED) {
+			return false;
+		}
+		struct array_literal **next = &v->literal.array;
+		for (size_t i = 0; i < t->array.length; i++) {
+			*next = xcalloc(1, sizeof(struct array_literal));
+			(*next)->value = xcalloc(1, sizeof(struct expression));
+			(*next)->value->type = EXPR_LITERAL;
+			(*next)->value->result = t->array.members;
+			if (!literal_default(ctx, (*next)->value)) {
+				return false;
+			}
+			next = &(*next)->next;
+		}
 		break;
 	case STORAGE_TAGGED:
 		return false;
@@ -797,8 +817,9 @@ literal_default(struct context *ctx, struct expression *v)
 	case STORAGE_OPAQUE:
 	case STORAGE_VALIST:
 		assert(0); // Invariant
-	case STORAGE_VOID:
 	case STORAGE_DONE:
+	case STORAGE_NOMEM:
+	case STORAGE_VOID:
 		break; // no-op
 	}
 
@@ -1031,6 +1052,113 @@ eval_tuple(struct context *ctx,
 	return true;
 }
 
+static bool
+eval_address_object(struct context *ctx,
+	const struct expression *restrict in,
+	struct expression *restrict out)
+{
+	const struct expression_access *access =
+		&in->unarithm.operand->access;
+	struct expression new_in = {0};
+	const struct type *operand_type;
+	switch (access->type) {
+	case ACCESS_IDENTIFIER:
+		if (access->object->otype != O_DECL) {
+			return false;
+		}
+		out->literal.object = access->object;
+		out->literal.ival = 0;
+		return true;
+	case ACCESS_INDEX:
+		new_in = *in;
+		new_in.unarithm.operand = access->array;
+		if (!eval_expr(ctx, &new_in, out)) {
+			return false;
+		}
+		struct expression index = {0};
+		if (!eval_expr(ctx, access->index, &index)) {
+			return false;
+		}
+		operand_type = type_dealias(ctx, access->array->result);
+		if (operand_type->storage != STORAGE_ARRAY) {
+			// autodereferencing not allowed
+			return false;
+		}
+		out->literal.ival +=
+			index.literal.uval * operand_type->array.members->size;
+		return true;
+	case ACCESS_FIELD:
+		new_in = *in;
+		new_in.unarithm.operand = access->_struct;
+		if (!eval_expr(ctx, &new_in, out)) {
+			return false;
+		}
+		operand_type = type_dealias(ctx, access->tuple->result);
+		if (operand_type->storage != STORAGE_STRUCT) {
+			// autodereferencing not allowed
+			return false;
+		}
+		out->literal.ival += access->field->offset;
+		return true;
+	case ACCESS_TUPLE:
+		new_in = *in;
+		new_in.unarithm.operand = access->tuple;
+		if (!eval_expr(ctx, &new_in, out)) {
+			return false;
+		}
+		operand_type = type_dealias(ctx, access->tuple->result);
+		if (operand_type->storage != STORAGE_TUPLE) {
+			// autodereferencing not allowed
+			return false;
+		}
+		out->literal.ival += access->tvalue->offset;
+		return true;
+	}
+	return true;
+}
+
+static bool
+eval_address_other(struct context *ctx,
+	const struct expression *restrict in,
+	struct expression *restrict out)
+{
+	struct expression *value = xcalloc(1, sizeof(struct expression));
+	if (!eval_expr(ctx, in->unarithm.operand, value)) {
+		return false;
+	}
+
+	static int serial = 0;
+	char *symbol = gen_name(&serial, "static.%d");
+	struct ident *name = mkident(ctx, NULL, symbol);
+
+	append_decl(ctx, &(struct declaration){
+		.decl_type = DECL_GLOBAL,
+		.file = in->loc.file,
+		.ident = name,
+		.symbol = symbol,
+		.exported = false,
+		.global = {
+			.type = value->result,
+			.value = value,
+			.threadlocal = false,
+		}
+	});
+
+	struct scope_object *obj = scope_insert(ctx->scope,
+			O_DECL, name, name, value->result, NULL);
+
+	struct expression shadow = *in;
+	shadow.unarithm.operand = &(struct expression){
+		.type = EXPR_ACCESS,
+		.access = (struct expression_access){
+			.type = ACCESS_IDENTIFIER,
+			.object = obj,
+		},
+	};
+	bool r = eval_address_object(ctx, &shadow, out);
+	assert(r);
+	return true;
+}
 
 static bool
 eval_unarithm(struct context *ctx,
@@ -1044,65 +1172,11 @@ eval_unarithm(struct context *ctx,
 			out->literal.uval = 0;
 			return true;
 		}
-		if (in->unarithm.operand->type != EXPR_ACCESS) {
-			return false;
-		}
-		const struct expression_access *access =
-			&in->unarithm.operand->access;
-		struct expression new_in = {0};
-		const struct type *operand_type;
-		switch (access->type) {
-		case ACCESS_IDENTIFIER:
-			if (access->object->otype != O_DECL) {
-				return false;
-			}
-			out->literal.object = access->object;
-			out->literal.ival = 0;
-			return true;
-		case ACCESS_INDEX:
-			new_in = *in;
-			new_in.unarithm.operand = access->array;
-			if (!eval_expr(ctx, &new_in, out)) {
-				return false;
-			}
-			struct expression index = {0};
-			if (!eval_expr(ctx, access->index, &index)) {
-				return false;
-			}
-			operand_type = type_dealias(ctx, access->array->result);
-			if (operand_type->storage != STORAGE_ARRAY) {
-				// autodereferencing not allowed
-				return false;
-			}
-			out->literal.ival +=
-				index.literal.uval * operand_type->array.members->size;
-			return true;
-		case ACCESS_FIELD:
-			new_in = *in;
-			new_in.unarithm.operand = access->_struct;
-			if (!eval_expr(ctx, &new_in, out)) {
-				return false;
-			}
-			operand_type = type_dealias(ctx, access->tuple->result);
-			if (operand_type->storage != STORAGE_STRUCT) {
-				// autodereferencing not allowed
-				return false;
-			}
-			out->literal.ival += access->field->offset;
-			return true;
-		case ACCESS_TUPLE:
-			new_in = *in;
-			new_in.unarithm.operand = access->tuple;
-			if (!eval_expr(ctx, &new_in, out)) {
-				return false;
-			}
-			operand_type = type_dealias(ctx, access->tuple->result);
-			if (operand_type->storage != STORAGE_TUPLE) {
-				// autodereferencing not allowed
-				return false;
-			}
-			out->literal.ival += access->tvalue->offset;
-			return true;
+		switch (in->unarithm.operand->type) {
+		case EXPR_ACCESS:
+			return eval_address_object(ctx, in, out);
+		default:
+			return eval_address_other(ctx, in, out);
 		}
 	}
 
@@ -1140,6 +1214,7 @@ eval_expr(struct context *ctx,
 	const struct expression *restrict in,
 	struct expression *restrict out)
 {
+	out->loc = in->loc;
 	out->result = in->result;
 	out->type = EXPR_LITERAL;
 
@@ -1157,7 +1232,7 @@ eval_expr(struct context *ctx,
 		case C_TEST:
 			return eval_type_test(ctx, in, out);
 		default:
-			assert(0); // Unreacheable
+			assert(0); // Unreachable
 		}
 	case EXPR_LEN:
 		return eval_len(ctx, in, out);
