@@ -15,7 +15,7 @@
 #include "utf8.h"
 #include "util.h"
 
-static noreturn void
+FORMAT(2, 3) static noreturn void
 error(struct location loc, const char *fmt, ...)
 {
 	xfprintf(stderr, "%s:%d:%d: ", sources[loc.file],
@@ -50,7 +50,7 @@ vsynerr(struct token *tok, va_list ap)
 		sources[tok->loc.file], tok->loc.lineno, tok->loc.colno);
 
 	while (t != T_EOF) {
-		if (t == T_NUMBER || t == T_NAME) {
+		if (t == T_LITERAL || t == T_NAME) {
 			xfprintf(stderr, "%s", lexical_token_str(t));
 		} else {
 			xfprintf(stderr, "'%s'", lexical_token_str(t));
@@ -94,9 +94,6 @@ want(struct lexer *lexer, enum lexical_token ltok, struct token *tok)
 	struct token *out = tok ? tok : &_tok;
 	lex(lexer, out);
 	synassert(out->token == ltok, out, ltok, T_EOF);
-	if (!tok) {
-		token_finish(out);
-	}
 }
 
 static struct ast_expression *
@@ -126,54 +123,43 @@ mkfuncparams(struct location loc)
 
 static struct ast_expression *parse_statement(struct lexer *lexer);
 
-bool
-parse_identifier(struct lexer *lexer, struct identifier *ident, bool trailing)
+struct ident *
+parse_identifier(struct lexer *lexer, const char *name, bool *trailing)
 {
-	struct token tok = {0};
-	struct location loc = {0};
-	struct identifier *i = ident;
-	*ident = (struct identifier){0};
+	struct token tok;
+	struct ident *i = &(struct ident){ .name = name, .ns = NULL };
 	size_t len = 0;
-	bool found_trailing = false;
-	while (!i->name) {
+	do {
 		switch (lex(lexer, &tok)) {
 		case T_NAME:
-			len += strlen(tok.name);
-			i->name = xstrdup(tok.name);
-			if (loc.file == 0) {
-				loc = tok.loc;
+			if (i->name) {
+				error(tok.loc, "unexpected %s", token_str(&tok));
 			}
-			token_finish(&tok);
+			len += strlen(tok.name);
+			i->name = tok.name;
 			break;
-		default:
-			synassert(trailing && i->ns, &tok, T_NAME, T_EOF);
-			unlex(lexer, &tok);
-			struct identifier *ns = i->ns;
-			*i = *ns;
-			free(ns);
-			found_trailing = true;
-			continue;
-		}
-
-		struct identifier *ns;
-		switch (lex(lexer, &tok)) {
 		case T_DOUBLE_COLON:
+			synassert(i->name, &tok, T_NAME, T_EOF);
 			len++;
-			ns = xcalloc(1, sizeof(struct identifier));
-			*ns = *i;
-			i->ns = ns;
+			i->ns = intern_ident(lexer->itbl, i->name, i->ns);
 			i->name = NULL;
 			break;
 		default:
 			unlex(lexer, &tok);
 			break;
 		}
+		if (len > IDENT_MAX) {
+			error(tok.loc, "identifier exceeds maximum length");
+		}
+	} while (tok.token == T_NAME || tok.token == T_DOUBLE_COLON);
+	if (trailing) {
+		*trailing = !i->name;
+		if (!i->name) {
+			return (struct ident *)i->ns;
+		}
 	}
-
-	if (len > IDENT_MAX) {
-		error(loc, "identifier exceeds maximum length");
-	}
-	return found_trailing;
+	synassert(i->name, &tok, T_NAME, T_EOF);
+	return intern_ident(lexer->itbl, i->name, i->ns);
 }
 
 static void
@@ -184,7 +170,7 @@ parse_import_members(struct lexer *lexer, struct ast_import_members **members)
 		*members = xcalloc(1, sizeof(struct ast_import_members));
 		want(lexer, T_NAME, &tok);
 		(*members)->loc = tok.loc;
-		(*members)->name = tok.name;
+		(*members)->name = intern_name(lexer->itbl, tok.name);
 		members = &(*members)->next;
 
 		switch (lex(lexer, &tok)) {
@@ -212,7 +198,8 @@ parse_import(struct lexer *lexer, struct ast_imports *import)
 {
 	struct token tok = {0};
 	import->mode = IMPORT_NORMAL;
-	bool trailing_colon = parse_identifier(lexer, &import->ident, true);
+	bool trailing_colon;
+	import->ident = parse_identifier(lexer, NULL, &trailing_colon);
 	if (trailing_colon) {
 		switch (lex(lexer, &tok)) {
 		case T_LBRACE:
@@ -225,12 +212,12 @@ parse_import(struct lexer *lexer, struct ast_imports *import)
 		default:
 			synerr(&tok, T_LBRACE, T_TIMES, T_EOF);
 		}
-	} else if (import->ident.ns == NULL) {
+	} else if (import->ident->ns == NULL) {
 		switch (lex(lexer, &tok)) {
 		case T_EQUAL:
 			import->mode = IMPORT_ALIAS;
-			import->alias = import->ident.name;
-			parse_identifier(lexer, &import->ident, false);
+			import->alias = import->ident->name;
+			import->ident = parse_identifier(lexer, NULL, NULL);
 			break;
 		default:
 			unlex(lexer, &tok);
@@ -273,29 +260,18 @@ parse_parameter_list(struct lexer *lexer, struct ast_function_type *type)
 	struct ast_function_parameters **next = &type->params;
 	for (;;) {
 		switch (lex(lexer, &tok)) {
-		case T_NAME:
-			// Assumes ownership of tok.name.
+		case T_NAME:;
 			switch (lex(lexer, &tok2)) {
 			case T_COLON:
-				(*next)->name = tok.name;
+				(*next)->name = intern_name(lexer->itbl, tok.name);
 				(*next)->type = parse_type(lexer);
 				break;
-			case T_DOUBLE_COLON:
-				(*next)->type = parse_type(lexer);
-				synassert((*next)->type->storage == STORAGE_ALIAS,
-						&tok, T_NAME, T_EOF);
-				struct identifier *ident =
-					xcalloc(1, sizeof(struct identifier));
-				struct identifier *ns;
-				ident->name = tok.name;
-				for (ns = &(*next)->type->alias; ns->ns; ns = ns->ns);
-				ns->ns = ident;
-				break;
 			default:
 				unlex(lexer, &tok2);
 				(*next)->type = mktype(tok.loc);
 				(*next)->type->storage = STORAGE_ALIAS;
-				(*next)->type->alias.name = tok.name;
+				(*next)->type->alias = parse_identifier(lexer,
+					tok.name, NULL);
 				break;
 			}
 			break;
@@ -424,6 +400,9 @@ parse_primitive_type(struct lexer *lexer)
 	case T_DONE:
 		type->storage = STORAGE_DONE;
 		break;
+	case T_NOMEM:
+		type->storage = STORAGE_NOMEM;
+		break;
 	case T_OPAQUE:
 		type->storage = STORAGE_OPAQUE;
 		break;
@@ -446,12 +425,12 @@ static struct ast_expression *parse_binding_list(
 static struct ast_expression *parse_object_selector(struct lexer *lexer);
 
 static struct ast_type *
-parse_enum_type(struct identifier *ident, struct lexer *lexer)
+parse_enum_type(struct ident *ident, struct lexer *lexer)
 {
 	struct token tok = {0};
 	struct ast_type *type = mktype(lexer->loc);
 	type->storage = STORAGE_ENUM;
-	identifier_dup(&type->alias, ident);
+	type->alias = ident;
 	struct ast_enum_field **next = &type->_enum.values;
 	switch (lex(lexer, &tok)) {
 	case T_LBRACE:
@@ -483,7 +462,7 @@ parse_enum_type(struct identifier *ident, struct lexer *lexer)
 	while (tok.token != T_RBRACE) {
 		*next = xcalloc(1, sizeof(struct ast_enum_field));
 		want(lexer, T_NAME, &tok);
-		(*next)->name = tok.name;
+		(*next)->name = intern_name(lexer->itbl, tok.name);
 		(*next)->loc = tok.loc;
 		if (lex(lexer, &tok) == T_EQUAL) {
 			(*next)->value = parse_expression(lexer);
@@ -538,34 +517,21 @@ parse_struct_union_type(struct lexer *lexer)
 			unlex(lexer, &tok);
 		}
 
-		char *name;
 		switch (lex(lexer, &tok)) {
-		case T_NAME:
-			name = tok.name;
+		case T_NAME:;
+			const char *name = tok.name;
 			struct location loc = tok.loc;
 			switch (lex(lexer, &tok)) {
 			case T_COLON:
 				next->name = name;
 				next->type = parse_type(lexer);
 				break;
-			case T_DOUBLE_COLON:
-				next->type = mktype(loc);
-				next->type->storage = STORAGE_ALIAS;
-				next->type->unwrap = false;
-				parse_identifier(lexer, &next->type->alias, false);
-				struct identifier *i = &next->type->alias;
-				while (i->ns != NULL) {
-					i = i->ns;
-				}
-				i->ns = xcalloc(1, sizeof(struct identifier));
-				i->ns->name = name;
-				break;
 			default:
 				unlex(lexer, &tok);
 				next->type = mktype(loc);
 				next->type->storage = STORAGE_ALIAS;
-				next->type->alias.name = name;
 				next->type->unwrap = false;
+				next->type->alias = parse_identifier(lexer, name, NULL);
 				break;
 			}
 			break;
@@ -671,13 +637,8 @@ parse_type(struct lexer *lexer)
 {
 	struct token tok = {0};
 	uint32_t flags = 0;
-	switch (lex(lexer, &tok)) {
-	case T_CONST:
-		flags |= TYPE_CONST;
-		break;
-	default:
+	if (lex(lexer, &tok) != T_CONST) {
 		unlex(lexer, &tok);
-		break;
 	}
 	switch (lex(lexer, &tok)) {
 	case T_LNOT:
@@ -691,6 +652,7 @@ parse_type(struct lexer *lexer)
 	bool nullable = false, unwrap = false;
 	switch (lex(lexer, &tok)) {
 	case T_BOOL:
+	case T_DONE:
 	case T_F32:
 	case T_F64:
 	case T_I16:
@@ -699,6 +661,7 @@ parse_type(struct lexer *lexer)
 	case T_I8:
 	case T_INT:
 	case T_NEVER:
+	case T_NOMEM:
 	case T_OPAQUE:
 	case T_RUNE:
 	case T_SIZE:
@@ -711,7 +674,6 @@ parse_type(struct lexer *lexer)
 	case T_UINTPTR:
 	case T_VALIST:
 	case T_VOID:
-	case T_DONE:
 		unlex(lexer, &tok);
 		type = parse_primitive_type(lexer);
 		break;
@@ -723,9 +685,7 @@ parse_type(struct lexer *lexer)
 		type = mktype(lexer->loc);
 		type->storage = STORAGE_POINTER;
 		type->pointer.referent = parse_type(lexer);
-		if (nullable) {
-			type->pointer.flags |= PTR_NULLABLE;
-		}
+		type->pointer.nullable = nullable;
 		break;
 	case T_STRUCT:
 	case T_UNION:
@@ -778,7 +738,7 @@ parse_type(struct lexer *lexer)
 		type = mktype(lexer->loc);
 		type->storage = STORAGE_ALIAS;
 		type->unwrap = unwrap;
-		parse_identifier(lexer, &type->alias, false);
+		type->alias = parse_identifier(lexer, NULL, NULL);
 		break;
 	default:
 		synassert_msg(false, "expected type", &tok);
@@ -790,7 +750,7 @@ parse_type(struct lexer *lexer)
 }
 
 static struct ast_expression *
-parse_access(struct lexer *lexer, struct identifier ident)
+parse_access(struct lexer *lexer, struct ident *ident)
 {
 	struct ast_expression *exp = mkexpr(lexer->loc);
 	exp->type = EXPR_ACCESS;
@@ -824,12 +784,15 @@ parse_literal(struct lexer *lexer)
 	case T_DONE:
 		exp->literal.storage = STORAGE_DONE;
 		return exp;
-	case T_NUMBER:
+	case T_NOMEM:
+		exp->literal.storage = STORAGE_NOMEM;
+		return exp;
+	case T_LITERAL:
 		exp->literal.storage = tok.storage;
 		break;
 	default:
-		synerr(&tok, T_NUMBER, T_TRUE,
-			T_FALSE, T_NULL, T_VOID, T_EOF);
+		synerr(&tok, T_LITERAL, T_TRUE, T_FALSE, T_NULL,
+			T_VOID, T_DONE, T_NOMEM, T_EOF);
 		break;
 	}
 
@@ -860,34 +823,28 @@ parse_literal(struct lexer *lexer)
 	case STORAGE_RCONST:
 		exp->literal.rune = tok.rune;
 		break;
-	case STORAGE_STRING:
-		exp->literal.string.len = tok.string.len;
-		exp->literal.string.value = tok.string.value;
-		while (lex(lexer, &tok) == T_NUMBER
-				&& tok.storage == STORAGE_STRING) {
-			size_t len = exp->literal.string.len;
-			exp->literal.string.value = xrealloc(
-				exp->literal.string.value,
-				len + tok.string.len);
-			memcpy(exp->literal.string.value + len,
-				tok.string.value, tok.string.len);
-			exp->literal.string.len += tok.string.len;
-		}
+	case STORAGE_STRING:;
+		size_t ln = 0, cap = 0;
+		char *s = NULL;
+		do {
+			append_buffer(&s, &ln, &cap, tok.string.value, tok.string.len);
+		} while (lex(lexer, &tok) == T_LITERAL && tok.storage == STORAGE_STRING);
 		unlex(lexer, &tok);
+		exp->literal.string.value = s;
+		exp->literal.string.len = ln;
 
 		// check for invalid UTF-8 (possible when \x is used)
-		const char *s = exp->literal.string.value;
-		size_t len = exp->literal.string.len;
-		while (s - exp->literal.string.value < (ptrdiff_t)len) {
-			if (utf8_decode(&s) == UTF8_INVALID) {
+		while (s - exp->literal.string.value < (ptrdiff_t)ln) {
+			if (utf8_decode((const char **)&s) == UTF8_INVALID) {
 				error(loc, "invalid UTF-8 in string literal");
 			}
 		}
 		break;
 	case STORAGE_BOOL:
+	case STORAGE_DONE:
+	case STORAGE_NOMEM:
 	case STORAGE_NULL:
 	case STORAGE_VOID:
-	case STORAGE_DONE:
 		assert(0); // Handled above
 	case STORAGE_ALIAS:
 	case STORAGE_ARRAY:
@@ -948,51 +905,35 @@ parse_array_literal(struct lexer *lexer)
 }
 
 static struct ast_expression *parse_struct_literal(struct lexer *lexer,
-	struct identifier ident);
+	struct ident *ident);
 
 static struct ast_field_value *
 parse_field_value(struct lexer *lexer)
 {
 	struct ast_field_value *exp =
 		xcalloc(1, sizeof(struct ast_field_value));
-	char *name;
 	struct token tok = {0};
-	struct identifier ident = {0};
-	struct identifier *i;
 	switch (lex(lexer, &tok)) {
-	case T_NAME:
-		name = tok.name;
+	case T_NAME:;
+		const char *name = tok.name;
 		switch (lex(lexer, &tok)) {
 		case T_COLON:
-			exp->name = name;
 			exp->type = parse_type(lexer);
 			want(lexer, T_EQUAL, NULL);
-			exp->initializer = parse_expression(lexer);
-			break;
+			/* fallthrough */
 		case T_EQUAL:
 			exp->name = name;
 			exp->initializer = parse_expression(lexer);
 			break;
-		case T_DOUBLE_COLON:
-			i = &ident;
-			parse_identifier(lexer, i, false);
-			while (i->ns != NULL) {
-				i = i->ns;
-			}
-			i->ns = xcalloc(1, sizeof(struct identifier));
-			i->ns->name = name;
-			exp->initializer = parse_struct_literal(lexer, ident);
-			break;
 		default:
 			unlex(lexer, &tok);
-			ident.name = name;
-			ident.ns = NULL;
-			exp->initializer = parse_struct_literal(lexer, ident);
+			struct ident *id = parse_identifier(lexer, name, NULL);
+			exp->initializer = parse_struct_literal(lexer, id);
 			break;
 		}
 		break;
 	case T_STRUCT:
-		exp->initializer = parse_struct_literal(lexer, ident);
+		exp->initializer = parse_struct_literal(lexer, NULL);
 		break;
 	default:
 		assert(0);
@@ -1001,7 +942,7 @@ parse_field_value(struct lexer *lexer)
 }
 
 static struct ast_expression *
-parse_struct_literal(struct lexer *lexer, struct identifier ident)
+parse_struct_literal(struct lexer *lexer, struct ident *ident)
 {
 	want(lexer, T_LBRACE, NULL);
 	struct ast_expression *exp = mkexpr(lexer->loc);
@@ -1012,7 +953,7 @@ parse_struct_literal(struct lexer *lexer, struct identifier ident)
 	while (tok.token != T_RBRACE) {
 		switch (lex(lexer, &tok)) {
 		case T_ELLIPSIS:
-			synassert(ident.name != NULL, &tok, T_RBRACE, T_EOF);
+			synassert(ident != NULL, &tok, T_RBRACE, T_EOF);
 			exp->_struct.autofill = true;
 			want(lexer, T_RBRACE, &tok);
 			unlex(lexer, &tok);
@@ -1088,34 +1029,32 @@ parse_plain_expression(struct lexer *lexer)
 	struct ast_expression *exp;
 	switch (lex(lexer, &tok)) {
 	// plain-expression
-	case T_NUMBER:
-	case T_TRUE:
-	case T_FALSE:
-	case T_NULL:
-	case T_VOID:
 	case T_DONE:
+	case T_FALSE:
+	case T_LITERAL:
+	case T_NOMEM:
+	case T_NULL:
+	case T_TRUE:
+	case T_VOID:
 		unlex(lexer, &tok);
 		return parse_literal(lexer);
 	case T_NAME:
 		unlex(lexer, &tok);
-		struct identifier ident = {0};
-		parse_identifier(lexer, &ident, false);
+		struct ident *id = parse_identifier(lexer, NULL, NULL);
 		switch (lex(lexer, &tok)) {
 		case T_LBRACE:
 			unlex(lexer, &tok);
-			return parse_struct_literal(lexer, ident);
+			return parse_struct_literal(lexer, id);
 		default:
 			unlex(lexer, &tok);
-			return parse_access(lexer, ident);
+			return parse_access(lexer, id);
 		}
 		assert(0);
 	case T_LBRACKET:
 		unlex(lexer, &tok);
 		return parse_array_literal(lexer);
 	case T_STRUCT:
-		ident.name = NULL;
-		ident.ns = NULL;
-		return parse_struct_literal(lexer, ident);
+		return parse_struct_literal(lexer, NULL);
 	// nested-expression
 	case T_LPAREN:
 		exp = parse_expression(lexer);
@@ -1128,11 +1067,8 @@ parse_plain_expression(struct lexer *lexer)
 			synerr(&tok, T_RPAREN, T_COMMA, T_EOF);
 		}
 		assert(0); // Unreachable
-	// empty block
-	case T_RBRACE:
-		error(tok.loc, "syntax error: cannot have empty block");
 	default:
-		synerr(&tok, T_NUMBER, T_NAME,
+		synerr(&tok, T_LITERAL, T_NAME,
 			T_LBRACKET, T_STRUCT, T_LPAREN, T_EOF);
 	}
 	assert(0); // Unreachable
@@ -1372,9 +1308,9 @@ parse_append_insert(struct lexer *lexer, struct location loc,
 
 	want(lexer, T_LPAREN, NULL);
 	expr->append.object = parse_object_selector(lexer);
-	if (etype == EXPR_INSERT) {
-		synassert_msg(expr->append.object->access.type == ACCESS_INDEX,
-				"expected indexing expression", &tok);
+	if (etype == EXPR_INSERT && expr->append.object->access.type != ACCESS_INDEX) {
+		error(expr->append.object->loc,
+			"syntax error: expected indexing expression");
 	}
 	want(lexer, T_COMMA, NULL);
 	expr->append.value = parse_expression(lexer);
@@ -1427,6 +1363,9 @@ parse_slice_mutation(struct lexer *lexer, bool is_static)
 	}
 }
 
+static struct ast_expression *
+parse_postfix_expression(struct lexer *lexer, struct ast_expression *lvalue);
+
 static struct ast_expression *
 parse_static_expression(struct lexer *lexer, bool allowbinding)
 {
@@ -1446,7 +1385,9 @@ parse_static_expression(struct lexer *lexer, bool allowbinding)
 	case T_INSERT:
 	case T_DELETE:
 		unlex(lexer, &tok);
-		return parse_slice_mutation(lexer, true);
+		struct ast_expression *lval = parse_slice_mutation(lexer, true);
+		// XXX: this is a hack around #954
+		return parse_postfix_expression(lexer, lval);
 	default:
 		if (allowbinding) {
 			synerr(&tok, T_LET, T_CONST, T_ABORT,
@@ -1459,11 +1400,82 @@ parse_static_expression(struct lexer *lexer, bool allowbinding)
 	assert(0); // Unreachable
 }
 
+static struct ast_expression *
+parse_va_expression(struct lexer *lexer)
+{
+	struct ast_expression *expr;
+	struct token tok;
+	switch (lex(lexer, &tok)) {
+	case T_VASTART:
+		expr = mkexpr(lexer->loc);
+		expr->type = EXPR_VASTART;
+		want(lexer, T_LPAREN, NULL);
+		want(lexer, T_RPAREN, NULL);
+		return expr;
+	case T_VAARG:
+		expr = mkexpr(lexer->loc);
+		expr->type = EXPR_VAARG;
+		want(lexer, T_LPAREN, NULL);
+		expr->vaarg.ap = parse_object_selector(lexer);
+		want(lexer, T_COMMA, NULL);
+		expr->vaarg.type = parse_type(lexer);
+		want(lexer, T_RPAREN, NULL);
+		return expr;
+	case T_VAEND:
+		expr = mkexpr(lexer->loc);
+		expr->type = EXPR_VAEND;
+		want(lexer, T_LPAREN, NULL);
+		expr->vaarg.ap = parse_object_selector(lexer);
+		want(lexer, T_RPAREN, NULL);
+		return expr;
+	default:
+		assert(0);
+	}
+}
+
+static struct ast_expression *
+parse_builtin_expression(struct lexer *lexer)
+{
+	struct token tok;
+	switch (lex(lexer, &tok)) {
+	case T_ALLOC:
+	case T_FREE:
+		unlex(lexer, &tok);
+		return parse_allocation_expression(lexer);
+	case T_APPEND:
+	case T_DELETE:
+	case T_INSERT:
+		unlex(lexer, &tok);
+		return parse_slice_mutation(lexer, false);
+	case T_STATIC:
+		return parse_static_expression(lexer, false);
+	case T_ABORT:
+	case T_ASSERT:
+		unlex(lexer, &tok);
+		return parse_assertion_expression(lexer, false);
+	case T_ALIGN:
+	case T_SIZE:
+	case T_LEN:
+	case T_OFFSET:
+		unlex(lexer, &tok);
+		return parse_measurement_expression(lexer);
+	case T_VAARG:
+	case T_VAEND:
+	case T_VASTART:
+		unlex(lexer, &tok);
+		return parse_va_expression(lexer);
+	default:
+		unlex(lexer, &tok);
+		break;
+	}
+	return parse_plain_expression(lexer);
+}
+
 static struct ast_expression *
 parse_postfix_expression(struct lexer *lexer, struct ast_expression *lvalue)
 {
 	if (lvalue == NULL) {
-		lvalue = parse_plain_expression(lexer);
+		lvalue = parse_builtin_expression(lexer);
 	}
 
 	struct token tok;
@@ -1483,14 +1495,14 @@ parse_postfix_expression(struct lexer *lexer, struct ast_expression *lvalue)
 			exp->access._struct = lvalue;
 			exp->access.field = tok.name;
 			break;
-		case T_NUMBER:
+		case T_LITERAL:
 			exp->access.type = ACCESS_TUPLE;
 			exp->access.tuple = lvalue;
 			unlex(lexer, &tok);
 			exp->access.value = parse_literal(lexer);
 			break;
 		default:
-			synerr(&tok, T_NAME, T_NUMBER, T_EOF);
+			synerr(&tok, T_NAME, T_LITERAL, T_EOF);
 		}
 
 		lvalue = exp;
@@ -1543,80 +1555,11 @@ parse_object_selector(struct lexer *lexer)
 	unlex(lexer, &tok);
 	struct ast_expression *exp = parse_postfix_expression(lexer, NULL);
 	synassert_msg(exp->type == EXPR_ACCESS,
-		"Expected object selector (identifier, indexing, or field access)",
+		"Expected object selector (ident, indexing, or field access)",
 		&tok);
 	return exp;
 }
 
-static struct ast_expression *
-parse_va_expression(struct lexer *lexer)
-{
-	struct ast_expression *expr;
-	struct token tok;
-	switch (lex(lexer, &tok)) {
-	case T_VASTART:
-		expr = mkexpr(lexer->loc);
-		expr->type = EXPR_VASTART;
-		want(lexer, T_LPAREN, NULL);
-		want(lexer, T_RPAREN, NULL);
-		return expr;
-	case T_VAARG:
-		expr = mkexpr(lexer->loc);
-		expr->type = EXPR_VAARG;
-		want(lexer, T_LPAREN, NULL);
-		expr->vaarg.ap = parse_object_selector(lexer);
-		want(lexer, T_RPAREN, NULL);
-		return expr;
-	case T_VAEND:
-		expr = mkexpr(lexer->loc);
-		expr->type = EXPR_VAEND;
-		want(lexer, T_LPAREN, NULL);
-		expr->vaarg.ap = parse_object_selector(lexer);
-		want(lexer, T_RPAREN, NULL);
-		return expr;
-	default:
-		assert(0);
-	}
-}
-
-static struct ast_expression *
-parse_builtin_expression(struct lexer *lexer)
-{
-	struct token tok;
-	switch (lex(lexer, &tok)) {
-	case T_ALLOC:
-	case T_FREE:
-		unlex(lexer, &tok);
-		return parse_allocation_expression(lexer);
-	case T_APPEND:
-	case T_DELETE:
-	case T_INSERT:
-		unlex(lexer, &tok);
-		return parse_slice_mutation(lexer, false);
-	case T_STATIC:
-		return parse_static_expression(lexer, false);
-	case T_ABORT:
-	case T_ASSERT:
-		unlex(lexer, &tok);
-		return parse_assertion_expression(lexer, false);
-	case T_ALIGN:
-	case T_SIZE:
-	case T_LEN:
-	case T_OFFSET:
-		unlex(lexer, &tok);
-		return parse_measurement_expression(lexer);
-	case T_VAARG:
-	case T_VAEND:
-	case T_VASTART:
-		unlex(lexer, &tok);
-		return parse_va_expression(lexer);
-	default:
-		unlex(lexer, &tok);
-		break;
-	}
-	return parse_postfix_expression(lexer, NULL);
-}
-
 static struct ast_expression *parse_compound_expression(struct lexer *lexer);
 static struct ast_expression *parse_match_expression(struct lexer *lexer);
 static struct ast_expression *parse_switch_expression(struct lexer *lexer);
@@ -1647,7 +1590,7 @@ parse_unary_expression(struct lexer *lexer)
 		return parse_switch_expression(lexer);
 	default:
 		unlex(lexer, &tok);
-		return parse_builtin_expression(lexer);
+		return parse_postfix_expression(lexer, NULL);
 	}
 }
 
@@ -1842,12 +1785,9 @@ static void parse_for_predicate(struct lexer *lexer,
 	struct ast_expression_for *for_exp)
 {
 	struct token tok = {0};
-	unsigned int flags = 0;
 
 	switch (lex(lexer, &tok)) {
 	case T_CONST:
-		flags = TYPE_CONST;
-		break;
 	case T_LET:
 		break;
 	default:
@@ -1864,11 +1804,9 @@ static void parse_for_predicate(struct lexer *lexer,
 
 		bool for_kind_found = false;
 		while (true) {
-			binding->flags = flags;
-
 			switch (lex(lexer, &tok)) {
 			case T_NAME:
-				binding->name = tok.name;
+				binding->name = intern_name(lexer->itbl, tok.name);
 				break;
 			case T_LPAREN:
 				parse_binding_unpack(lexer, &binding->unpack);
@@ -1879,7 +1817,6 @@ static void parse_for_predicate(struct lexer *lexer,
 
 			if (lex(lexer, &tok) == T_COLON) {
 				binding->type = parse_type(lexer);
-				binding->type->flags |= flags;
 			} else {
 				unlex(lexer, &tok);
 			}
@@ -2033,8 +1970,18 @@ parse_switch_expression(struct lexer *lexer)
 		struct ast_switch_case *_case =
 			*next_case = xcalloc(1, sizeof(struct ast_switch_case));
 		want(lexer, T_CASE, &tok);
+		struct location caseloc = tok.loc;
 		_case->options = parse_case_options(lexer);
 
+		switch (lex(lexer, &tok)) {
+		case T_CASE:
+		case T_RBRACE:
+			error(caseloc, "syntax error: case cannot be empty");
+		default:
+			unlex(lexer, &tok);
+			break;
+		}
+
 		bool exprs = true;
 		struct ast_expression_list *cur = &_case->exprs;
 		struct ast_expression_list **next = &cur->next;
@@ -2104,12 +2051,14 @@ parse_match_expression(struct lexer *lexer)
 		struct ast_match_case *_case =
 			*next_case = xcalloc(1, sizeof(struct ast_match_case));
 		want(lexer, T_CASE, &tok);
+		struct location caseloc = tok.loc;
 
+		_case->name = NULL;
 		struct ast_type *type = NULL;
 		switch (lex(lexer, &tok)) {
 		case T_LET:
 			want(lexer, T_NAME, &tok);
-			_case->name = tok.name;
+			_case->name = intern_name(lexer->itbl, tok.name);
 			want(lexer, T_COLON, NULL);
 			_case->type = parse_type(lexer);
 			break;
@@ -2129,6 +2078,14 @@ parse_match_expression(struct lexer *lexer)
 		}
 
 		want(lexer, T_ARROW, &tok);
+		switch (lex(lexer, &tok)) {
+		case T_CASE:
+		case T_RBRACE:
+			error(caseloc, "syntax error: case cannot be empty");
+		default:
+			unlex(lexer, &tok);
+			break;
+		}
 
 		bool exprs = true;
 		struct ast_expression_list *cur = &_case->exprs;
@@ -2173,14 +2130,15 @@ parse_match_expression(struct lexer *lexer)
 static void
 parse_binding_unpack(struct lexer *lexer, struct ast_binding_unpack **next)
 {
-	struct token tok = {0};
+	struct token tok;
+	while (true) {
+		struct ast_binding_unpack *new = xcalloc(1, sizeof *new);
+		*next = new;
+		next = &new->next;
 
-	bool more = true;
-	while (more) {
-		char *name = NULL;
 		switch (lex(lexer, &tok)) {
 		case T_NAME:
-			name = tok.name;
+			new->name = intern_name(lexer->itbl, tok.name);
 			break;
 		case T_UNDERSCORE:
 			break;
@@ -2188,18 +2146,11 @@ parse_binding_unpack(struct lexer *lexer, struct ast_binding_unpack **next)
 			synerr(&tok, T_NAME, T_UNDERSCORE, T_EOF);
 		}
 
-		struct ast_binding_unpack *new = xcalloc(1, sizeof *new);
-		*next = new;
-		next = &new->next;
-
-		new->name = name;
-
 		switch (lex(lexer, &tok)) {
 		case T_COMMA:
 			break;
 		case T_RPAREN:
-			more = false;
-			break;
+			return;
 		default:
 			synerr(&tok, T_COMMA, T_RPAREN, T_EOF);
 		}
@@ -2210,7 +2161,6 @@ static struct ast_expression *
 parse_binding_list(struct lexer *lexer, bool is_static)
 {
 	struct ast_expression *exp = mkexpr(lexer->loc);
-	unsigned int flags = 0;
 
 	struct token tok = {0};
 	switch (lex(lexer, &tok)) {
@@ -2218,8 +2168,6 @@ parse_binding_list(struct lexer *lexer, bool is_static)
 		exp->type = EXPR_DEFINE;
 		break;
 	case T_CONST:
-		flags = TYPE_CONST;
-		// fallthrough
 	case T_LET:
 		exp->type = EXPR_BINDING;
 		break;
@@ -2228,31 +2176,24 @@ parse_binding_list(struct lexer *lexer, bool is_static)
 	}
 
 	struct ast_expression_binding *binding = &exp->binding;
-	struct ast_expression_binding **next = &exp->binding.next;
 
-	bool more = true;
-	while (more) {
+	do {
 		switch (lex(lexer, &tok)) {
 		case T_NAME:
-			binding->name = tok.name;
+			binding->name = intern_name(lexer->itbl, tok.name);
 			break;
 		case T_LPAREN:
-			if (exp->type == EXPR_BINDING) {
-				parse_binding_unpack(lexer, &binding->unpack);
-				break;
-			}
-			// fallthrough
+			synassert(exp->type == EXPR_BINDING, &tok, T_NAME, T_EOF);
+			parse_binding_unpack(lexer, &binding->unpack);
+			break;
 		default:
 			synerr(&tok, T_NAME, T_LPAREN, T_EOF);
 		}
-		binding->initializer = mkexpr(lexer->loc);
-		binding->flags = flags;
 		binding->is_static = is_static;
 
 		switch (lex(lexer, &tok)) {
 		case T_COLON:
 			binding->type = parse_type(lexer);
-			binding->type->flags |= flags;
 			want(lexer, T_EQUAL, &tok);
 			break;
 		case T_EQUAL:
@@ -2263,18 +2204,13 @@ parse_binding_list(struct lexer *lexer, bool is_static)
 
 		binding->initializer = parse_expression(lexer);
 
-		switch (lex(lexer, &tok)) {
-		case T_COMMA:
-			*next = xcalloc(1, sizeof(struct ast_expression_binding));
-			binding = *next;
-			next = &binding->next;
-			break;
-		default:
-			unlex(lexer, &tok);
-			more = false;
-			break;
+		if (lex(lexer, &tok) == T_COMMA) {
+			binding->next = xcalloc(1, sizeof *binding->next);
+			binding = binding->next;
 		}
-	}
+	} while (tok.token == T_COMMA);
+
+	unlex(lexer, &tok);
 	return exp;
 }
 
@@ -2399,6 +2335,15 @@ parse_compound_expression(struct lexer *lexer)
 		break;
 	}
 
+	struct location loc = tok.loc;
+	switch (lex(lexer, &tok)) {
+	case T_RBRACE:
+		error(loc, "syntax error: cannot have empty block");
+	default:
+		unlex(lexer, &tok);
+		break;
+	}
+
 	while (true) {
 		cur->expr = parse_statement(lexer);
 
@@ -2506,24 +2451,29 @@ parse_statement(struct lexer *lexer)
 	}
 }
 
-static char *
+static const char *
 parse_attr_symbol(struct lexer *lexer)
 {
-	struct token tok = {0};
 	want(lexer, T_LPAREN, NULL);
-	want(lexer, T_NUMBER, &tok);
-	synassert_msg(tok.storage == STORAGE_STRING,
-		"expected string literal", &tok);
-	synassert_msg(tok.string.len > 0, "invalid symbol", &tok);
-	for (size_t i = 0; i < tok.string.len; i++) {
-		uint32_t c = tok.string.value[i];
-		synassert_msg(c <= 0x7F && (isalnum(c) || c == '_' || c == '$'
-			|| c == '.'), "invalid symbol", &tok);
-		synassert_msg(i != 0 || (!isdigit(c) && c != '$'),
-			"invalid symbol", &tok);
-	}
+	struct ast_expression *exp = parse_literal(lexer);
 	want(lexer, T_RPAREN, NULL);
-	return tok.string.value;
+
+	if (exp->type != EXPR_LITERAL || exp->literal.storage != STORAGE_STRING
+			|| exp->literal.string.len == 0) {
+		error(exp->loc, "expected nonempty string literal");
+	}
+	char *s = exp->literal.string.value;
+	if ((uint32_t)s[0] > 0x7F || isdigit(s[0]) || s[0] == '$') {
+		error(exp->loc, "invalid symbol %s", s);
+	}
+	for (size_t i = 0; i < exp->literal.string.len; i++) {
+		uint32_t c = s[i];
+		if (c > 0x7F || !(isalnum(c) || c == '_' || c == '$' || c == '.')) {
+			error(exp->loc, "invalid symbol %s", s);
+		}
+	}
+	free(exp);
+	return intern_owned(lexer->itbl, s);
 }
 
 static void
@@ -2553,13 +2503,10 @@ parse_global_decl(struct lexer *lexer, enum lexical_token mode,
 				break;
 			}
 		}
-		parse_identifier(lexer, &i->ident, false);
+		i->ident = parse_identifier(lexer, NULL, NULL);
 		switch (lex(lexer, &tok)) {
 		case T_COLON:
 			i->type = parse_type(lexer);
-			if (mode == T_CONST) {
-				i->type->flags |= TYPE_CONST;
-			}
 			if (lex(lexer, &tok) != T_EQUAL) {
 				synassert(mode != T_DEF, &tok, T_EQUAL, T_EOF);
 				unlex(lexer, &tok);
@@ -2599,11 +2546,11 @@ parse_type_decl(struct lexer *lexer, struct ast_type_decl *decl)
 	struct ast_type_decl *i = decl;
 	bool more = true;
 	while (more) {
-		parse_identifier(lexer, &i->ident, false);
+		i->ident = parse_identifier(lexer, NULL, NULL);
 		want(lexer, T_EQUAL, NULL);
 		switch (lex(lexer, &tok)) {
 		case T_ENUM:
-			i->type = parse_enum_type(&i->ident, lexer);
+			i->type = parse_enum_type(i->ident, lexer);
 			break;
 		default:
 			unlex(lexer, &tok);
@@ -2652,7 +2599,7 @@ parse_fn_decl(struct lexer *lexer, struct ast_function_decl *decl)
 		}
 	}
 	want(lexer, T_FN, NULL);
-	parse_identifier(lexer, &decl->ident, false);
+	decl->ident = parse_identifier(lexer, NULL, NULL);
 	parse_prototype(lexer, &decl->prototype);
 
 	switch (lex(lexer, &tok)) {
