@@ -4,16 +4,17 @@
 #include <stdio.h>
 #include "check.h"
 #include "eval.h"
+#include "identifier.h"
 #include "scope.h"
 #include "type_store.h"
 #include "types.h"
 #include "util.h"
 
-static struct dimensions lookup_atype_with_dimensions(struct context *ctx,
-		const struct type **type, const struct ast_type *atype);
-
-static const struct type *
-lookup_atype(struct context *ctx, const struct ast_type *atype);
+static struct dimensions
+dim_from_type(const struct type *type)
+{
+	return (struct dimensions){ .size = type->size, .align = type->align };
+}
 
 static size_t
 ast_array_len(struct context *ctx, const struct ast_type *atype)
@@ -42,57 +43,59 @@ ast_array_len(struct context *ctx, const struct ast_type *atype)
 }
 
 const struct type *
-builtin_type_for_storage(enum type_storage storage, bool is_const)
+builtin_type_for_storage(enum type_storage storage)
 {
 	switch (storage) {
 	case STORAGE_BOOL:
-		return is_const ? &builtin_type_const_bool : &builtin_type_bool;
+		return &builtin_type_bool;
 	case STORAGE_ERROR:
 		return &builtin_type_error;
 	case STORAGE_F32:
-		return is_const ? &builtin_type_const_f32 : &builtin_type_f32;
+		return &builtin_type_f32;
 	case STORAGE_F64:
-		return is_const ? &builtin_type_const_f64 : &builtin_type_f64;
+		return &builtin_type_f64;
 	case STORAGE_I8:
-		return is_const ? &builtin_type_const_i8 : &builtin_type_i8;
+		return &builtin_type_i8;
 	case STORAGE_I16:
-		return is_const ? &builtin_type_const_i16 : &builtin_type_i16;
+		return &builtin_type_i16;
 	case STORAGE_I32:
-		return is_const ? &builtin_type_const_i32 : &builtin_type_i32;
+		return &builtin_type_i32;
 	case STORAGE_I64:
-		return is_const ? &builtin_type_const_i64 : &builtin_type_i64;
+		return &builtin_type_i64;
 	case STORAGE_INT:
-		return is_const ? &builtin_type_const_int : &builtin_type_int;
+		return &builtin_type_int;
 	case STORAGE_NEVER:
-		return is_const ? &builtin_type_const_never : &builtin_type_never;
+		return &builtin_type_never;
+	case STORAGE_NOMEM:
+		return &builtin_type_nomem;
 	case STORAGE_OPAQUE:
-		return is_const ? &builtin_type_const_opaque : &builtin_type_opaque;
+		return &builtin_type_opaque;
 	case STORAGE_RUNE:
-		return is_const ? &builtin_type_const_rune : &builtin_type_rune;
+		return &builtin_type_rune;
 	case STORAGE_SIZE:
-		return is_const ? &builtin_type_const_size : &builtin_type_size;
+		return &builtin_type_size;
 	case STORAGE_U8:
-		return is_const ? &builtin_type_const_u8 : &builtin_type_u8;
+		return &builtin_type_u8;
 	case STORAGE_U16:
-		return is_const ? &builtin_type_const_u16 : &builtin_type_u16;
+		return &builtin_type_u16;
 	case STORAGE_U32:
-		return is_const ? &builtin_type_const_u32 : &builtin_type_u32;
+		return &builtin_type_u32;
 	case STORAGE_U64:
-		return is_const ? &builtin_type_const_u64 : &builtin_type_u64;
+		return &builtin_type_u64;
 	case STORAGE_UINT:
-		return is_const ? &builtin_type_const_uint : &builtin_type_uint;
+		return &builtin_type_uint;
 	case STORAGE_UINTPTR:
-		return is_const ? &builtin_type_const_uintptr : &builtin_type_uintptr;
+		return &builtin_type_uintptr;
 	case STORAGE_VALIST:
 		return &builtin_type_valist;
 	case STORAGE_VOID:
-		return is_const ? &builtin_type_const_void : &builtin_type_void;
+		return &builtin_type_void;
 	case STORAGE_DONE:
-		return is_const ? &builtin_type_const_done : &builtin_type_done;
+		return &builtin_type_done;
 	case STORAGE_NULL:
-		return &builtin_type_null; // const null and null are the same type
+		return &builtin_type_null;
 	case STORAGE_STRING:
-		return is_const ? &builtin_type_const_str : &builtin_type_str;
+		return &builtin_type_str;
 	case STORAGE_ALIAS:
 	case STORAGE_ARRAY:
 	case STORAGE_FUNCTION:
@@ -117,8 +120,7 @@ builtin_for_type(const struct type *type)
 	if (type->flags & TYPE_ERROR) {
 		return NULL;
 	}
-	bool is_const = (type->flags & TYPE_CONST) != 0;
-	return builtin_type_for_storage(type->storage, is_const);
+	return builtin_type_for_storage(type->storage);
 }
 
 static bool
@@ -160,13 +162,14 @@ struct_new_field(struct context *ctx, struct type *type,
 	struct struct_field *field = xcalloc(1, sizeof(struct struct_field));
 
 	if (afield->name && !size_only) {
-		field->name = xstrdup(afield->name);
+		field->name = afield->name;
 	}
 	struct dimensions dim = {0};
 	if (size_only) {
-		dim = lookup_atype_with_dimensions(ctx, NULL, afield->type);
+		dim = type_store_lookup_dimensions(ctx, afield->type);
 	} else {
-		dim = lookup_atype_with_dimensions(ctx, &field->type, afield->type);
+		field->type = type_store_lookup_atype(ctx, afield->type);
+		dim = dim_from_type(field->type);
 	}
 	if (afield->next != NULL && dim.size == SIZE_UNDEFINED) {
 		error(ctx, afield->type->loc, NULL,
@@ -243,9 +246,10 @@ struct_new_field(struct context *ctx, struct type *type,
 	return field;
 }
 
-static const struct type *type_store_lookup_type(struct context *ctx, const struct type *type);
+static const struct type *type_store_lookup_type(struct context *ctx,
+		const struct type *type);
 
-bool
+static bool
 check_embedded_member(struct context *ctx,
 	const struct ast_struct_union_field *afield,
 	struct struct_field *member,
@@ -281,7 +285,7 @@ check_embedded_member(struct context *ctx,
 	return true;
 }
 
-void
+static void
 shift_fields(struct context *ctx,
 	const struct ast_struct_union_field *afield, struct struct_field *parent)
 {
@@ -312,7 +316,7 @@ shift_fields(struct context *ctx,
 		new->type = field->type;
 		new->offset = parent->offset;
 		if (field->name) {
-			new->name = xstrdup(field->name);
+			new->name = field->name;
 		} else {
 			shift_fields(ctx, NULL, new);
 		}
@@ -398,8 +402,7 @@ enforce_tagged_invariants(struct context *ctx, struct location loc,
 }
 
 static size_t
-sum_tagged_memb(struct context *ctx,
-		const struct type_tagged_union *u)
+sum_tagged_memb(struct context *ctx, const struct type_tagged_union *u)
 {
 	size_t nmemb = 0;
 	for (; u; u = u->next) {
@@ -421,9 +424,9 @@ tagged_or_atagged_member(struct context *ctx,
 	const struct ast_type *_atype = *atype;
 	while (_atype->storage == STORAGE_ALIAS && _atype->unwrap) {
 		const struct scope_object *obj = scope_lookup(
-			ctx->scope, &_atype->alias);
+			ctx->scope, _atype->alias);
 		if (!obj) {
-			char *ident = identifier_unparse(&_atype->alias);
+			char *ident = ident_unparse(_atype->alias);
 			error(ctx, _atype->loc, NULL,
 				"Unknown object '%s'", ident);
 			free(ident);
@@ -435,7 +438,7 @@ tagged_or_atagged_member(struct context *ctx,
 				*type = type_dealias(ctx, obj->type);
 				return;
 			} else {
-				char *ident = identifier_unparse(&obj->ident);
+				char *ident = ident_unparse(obj->ident);
 				error(ctx, _atype->loc, NULL,
 					"Object '%s' is not a type", ident);
 				free(ident);
@@ -443,18 +446,16 @@ tagged_or_atagged_member(struct context *ctx,
 				return;
 			}
 		}
-		struct incomplete_declaration *idecl =
-			(struct incomplete_declaration *)obj;
-		if (idecl->type != IDECL_DECL
-				|| idecl->decl.decl_type != ADECL_TYPE) {
-			char *ident = identifier_unparse(&obj->ident);
+		if (obj->idecl->type != IDECL_DECL
+				|| obj->idecl->decl.decl_type != ADECL_TYPE) {
+			char *ident = ident_unparse(obj->ident);
 			error(ctx, _atype->loc, NULL,
 				"Object '%s' is not a type", ident);
 			free(ident);
 			*type = &builtin_type_error;
 			return;
 		}
-		_atype = idecl->decl.type.type;
+		_atype = obj->idecl->decl.type.type;
 	}
 	*type = NULL;
 	*atype = _atype;
@@ -505,7 +506,7 @@ collect_atagged_memb(struct context *ctx,
 		size_t *i)
 {
 	for (; atu; atu = atu->next) {
-		const struct type *type = lookup_atype(ctx, atu->type);
+		const struct type *type = type_store_lookup_atype(ctx, atu->type);
 		if (type->storage == STORAGE_TAGGED) {
 			collect_tagged_memb(ctx, ta, &type->tagged, i);
 			continue;
@@ -606,7 +607,7 @@ _tagged_size(struct context *ctx, const struct ast_tagged_union_type *u)
 		} else if (atype->storage == STORAGE_TAGGED) {
 			memb = _tagged_size(ctx, &atype->tagged);
 		} else {
-			memb = lookup_atype_with_dimensions(ctx, NULL, atype);
+			memb = type_store_lookup_dimensions(ctx, atype);
 		}
 		if (memb.size == SIZE_UNDEFINED) {
 			error(ctx, atype->loc, NULL,
@@ -649,11 +650,11 @@ tuple_init_from_atype(struct context *ctx,
 	struct dimensions dim = {0};
 	while (atuple) {
 		struct dimensions memb = {0};
-		size_t offset = 0;
 		if (type) {
-			memb = lookup_atype_with_dimensions(ctx, &cur->type, atuple->type);
+			cur->type = type_store_lookup_atype(ctx, atuple->type);
+			memb = dim_from_type(cur->type);
 		} else {
-			memb = lookup_atype_with_dimensions(ctx, NULL, atuple->type);
+			memb = type_store_lookup_dimensions(ctx, atuple->type);
 		}
 		if (memb.size == SIZE_UNDEFINED) {
 			error(ctx, atype->loc, NULL,
@@ -663,8 +664,8 @@ tuple_init_from_atype(struct context *ctx,
 			}
 			return (struct dimensions){0};
 		}
+		size_t offset = dim.size;
 		if (memb.align != 0) {
-			offset = dim.size;
 			if (dim.size % memb.align) {
 				offset += memb.align - dim.size % memb.align;
 			}
@@ -676,9 +677,7 @@ tuple_init_from_atype(struct context *ctx,
 
 		atuple = atuple->next;
 		if (type) {
-			if (memb.align != 0) {
-				cur->offset = offset;
-			}
+			cur->offset = offset;
 			if (atuple) {
 				cur->next = xcalloc(1, sizeof(struct type_tuple));
 				cur = cur->next;
@@ -709,6 +708,9 @@ default_param_from_atype(struct context *ctx,
 	// updated later, so it cannot be on the stack.
 	struct expression *in = xcalloc(1, sizeof(struct expression));
 	check_expression(ctx, aparam->default_value, in, param->type);
+	if (in->result->storage == STORAGE_ERROR) {
+		return false;
+	}
 	if (!type_is_assignable(ctx, param->type, in->result)) {
 		char *restypename = gen_typename(in->result);
 		char *partypename = gen_typename(param->type);
@@ -761,6 +763,16 @@ type_init_from_atype(struct context *ctx,
 	case STORAGE_ENUM:
 	case STORAGE_NULL:
 		assert(0); // Invariant
+	case STORAGE_DONE:
+	case STORAGE_NEVER:
+		if (atype->flags & TYPE_ERROR) {
+			error(ctx, atype->loc, NULL,
+				"Error flag can't be used on %s type",
+				type_storage_unparse(atype->storage));
+			*type = builtin_type_error;
+			return (struct dimensions){0};
+		}
+		// fallthrough
 	case STORAGE_BOOL:
 	case STORAGE_F32:
 	case STORAGE_F64:
@@ -769,7 +781,7 @@ type_init_from_atype(struct context *ctx,
 	case STORAGE_I32:
 	case STORAGE_I64:
 	case STORAGE_INT:
-	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_RUNE:
 	case STORAGE_SIZE:
@@ -782,15 +794,14 @@ type_init_from_atype(struct context *ctx,
 	case STORAGE_UINTPTR:
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
-	case STORAGE_DONE:
-		builtin = builtin_type_for_storage(type->storage, false);
+		builtin = builtin_type_for_storage(type->storage);
 		type->size = builtin->size;
 		type->align = builtin->align;
 		break;
 	case STORAGE_ALIAS:
-		obj = scope_lookup(ctx->scope, &atype->alias);
+		obj = scope_lookup(ctx->scope, atype->alias);
 		if (!obj) {
-			char *ident = identifier_unparse(&atype->alias);
+			char *ident = ident_unparse(atype->alias);
 			error(ctx, atype->loc, NULL,
 				"Unresolvable identifier '%s'", ident);
 			free(ident);
@@ -800,9 +811,7 @@ type_init_from_atype(struct context *ctx,
 
 		if (obj->otype == O_SCAN) {
 			// an incomplete declaration was encountered
-			struct incomplete_declaration *idecl =
-				(struct incomplete_declaration *)obj;
-			if (size_only && idecl->type == IDECL_DECL) {
+			if (size_only && obj->idecl->type == IDECL_DECL) {
 				wrap_resolver(ctx, obj, resolve_dimensions);
 				type->size = obj->type->size;
 				type->align = obj->type->align;
@@ -813,7 +822,7 @@ type_init_from_atype(struct context *ctx,
 		}
 
 		if (obj->otype != O_TYPE) {
-			char *ident = identifier_unparse(&obj->ident);
+			char *ident = ident_unparse(obj->ident);
 			error(ctx, atype->loc, NULL,
 				"Object '%s' is not a type", ident);
 			free(ident);
@@ -828,8 +837,14 @@ type_init_from_atype(struct context *ctx,
 			*type = *type_dealias(ctx, obj->type);
 			break;
 		}
-		identifier_dup(&type->alias.ident, &obj->ident);
-		identifier_dup(&type->alias.name, &obj->name);
+		if ((atype->flags & TYPE_ERROR) && type_is_done(ctx, obj->type)) {
+			error(ctx, atype->loc, NULL,
+				"Error flag can't be used on done type");
+			*type = builtin_type_error;
+			return (struct dimensions){0};
+		}
+		type->alias.ident = obj->ident;
+		type->alias.name = obj->name;
 		type->alias.type = obj->type->alias.type;
 		type->alias.exported = obj->type->alias.exported;
 		type->size = obj->type->size;
@@ -839,11 +854,12 @@ type_init_from_atype(struct context *ctx,
 		type->array.length = ast_array_len(ctx, atype);
 		struct dimensions memb = {0};
 		if (size_only) {
-			memb = lookup_atype_with_dimensions(ctx,
-				NULL, atype->array.members);
+			memb = type_store_lookup_dimensions(ctx,
+				atype->array.members);
 		} else {
-			memb = lookup_atype_with_dimensions(ctx,
-				&type->array.members, atype->array.members);
+			type->array.members = type_store_lookup_atype(ctx,
+				atype->array.members);
+			memb = dim_from_type(type->array.members);
 			if (type->array.members->storage == STORAGE_ERROR) {
 				*type = builtin_type_error;
 				return (struct dimensions){0};
@@ -875,14 +891,15 @@ type_init_from_atype(struct context *ctx,
 		if (size_only) {
 			break;
 		}
-		type->func.result = lookup_atype(ctx, atype->func.result);
+		type->func.result = type_store_lookup_atype(ctx,
+				atype->func.result);
 		type->func.variadism = atype->func.variadism;
 		struct type_func_param *param, **next = &type->func.params;
 		bool has_optional = false;
 		for (struct ast_function_parameters *aparam = atype->func.params;
 				aparam; aparam = aparam->next) {
 			param = *next = xcalloc(1, sizeof(struct type_func_param));
-			param->type = lookup_atype(ctx, aparam->type);
+			param->type = type_store_lookup_atype(ctx, aparam->type);
 			if (param->type->size == SIZE_UNDEFINED) {
 				error(ctx, atype->loc, NULL,
 					"Function parameter types must have defined size");
@@ -915,8 +932,8 @@ type_init_from_atype(struct context *ctx,
 		if (size_only) {
 			break;
 		}
-		type->pointer.flags = atype->pointer.flags;
-		type->pointer.referent = lookup_atype(
+		type->pointer.nullable = atype->pointer.nullable;
+		type->pointer.referent = type_store_lookup_atype(
 			ctx, atype->pointer.referent);
 		if (type->pointer.referent->storage == STORAGE_ERROR) {
 			*type = builtin_type_error;
@@ -942,7 +959,8 @@ type_init_from_atype(struct context *ctx,
 		if (size_only) {
 			break;
 		}
-		type->array.members = lookup_atype(ctx, atype->slice.members);
+		type->array.members = type_store_lookup_atype(ctx,
+				atype->slice.members);
 		if (type->array.members->storage == STORAGE_ERROR) {
 			*type = builtin_type_error;
 			return (struct dimensions){0};
@@ -955,7 +973,7 @@ type_init_from_atype(struct context *ctx,
 		}
 		if (type->array.members->storage == STORAGE_NEVER) {
 			error(ctx, atype->loc, NULL,
-				"Never is not a valid slice member");
+				"never is not a valid slice member");
 			*type = builtin_type_error;
 			return (struct dimensions){0};
 		}
@@ -967,6 +985,9 @@ type_init_from_atype(struct context *ctx,
 			*type = builtin_type_error;
 			return (struct dimensions){0};
 		}
+		if (type->storage == STORAGE_UNION || !type->struct_union.packed) {
+			add_padding(&type->size, type->align);
+		}
 		break;
 	case STORAGE_TAGGED:
 		if (size_only) {
@@ -976,6 +997,7 @@ type_init_from_atype(struct context *ctx,
 		} else {
 			tagged_init_from_atype(ctx, type, atype);
 		}
+		add_padding(&type->size, type->align);
 		break;
 	case STORAGE_TUPLE:
 		if (size_only) {
@@ -986,27 +1008,14 @@ type_init_from_atype(struct context *ctx,
 		} else {
 			tuple_init_from_atype(ctx, type, atype);
 		}
+		add_padding(&type->size, type->align);
 		break;
 	}
-
-	struct dimensions dim = {
-		.size = type->size,
-		.align = type->align,
-	};
-	if (type->storage != STORAGE_STRUCT || !type->struct_union.packed) {
-		// padding an alias can only break packed structs
-		if (type->storage != STORAGE_ALIAS) {
-			add_padding(&dim.size, dim.align);
-		}
-	}
-	return dim;
+	return dim_from_type(type);
 }
 
 static const struct type *
-_type_store_lookup_type(
-	struct context *ctx,
-	const struct type *type,
-	const struct dimensions *dims)
+type_store_lookup_type(struct context *ctx, const struct type *type)
 {
 	const struct type *builtin = builtin_for_type(type);
 	if (builtin) {
@@ -1035,60 +1044,25 @@ _type_store_lookup_type(
 	bucket = *next = xcalloc(1, sizeof(struct type_bucket));
 	bucket->type = *type;
 	bucket->type.id = hash;
-
-	if (dims == NULL) {
-		add_padding(&bucket->type.size, type->align);
-	}
-
 	return &bucket->type;
 }
 
-static const struct type *
-type_store_lookup_type(struct context *ctx, const struct type *type)
-{
-	if (type->storage != STORAGE_ALIAS) {
-		return _type_store_lookup_type(ctx, type, NULL);
-	}
-	// References to type aliases always inherit the flags that the
-	// alias was defined with
-	struct type psuedotype = *type;
-	const struct scope_object *obj = scope_lookup(
-		ctx->scope, &type->alias.name);
-	psuedotype.flags |= obj->type->flags;
-	return type_store_lookup_alias(ctx, &psuedotype, NULL);
-}
-
-static struct dimensions
-lookup_atype_with_dimensions(struct context *ctx,
-	const struct type **type,
-	const struct ast_type *atype)
-{
-	struct type temp = {0};
-	struct dimensions dim = {0};
-	if (type) {
-		dim = type_init_from_atype(ctx, &temp, atype);
-		*type = type_store_lookup_type(ctx, &temp);
-	} else {
-		dim = type_init_from_atype(ctx, NULL, atype);
-	}
-	return dim;
-}
-
-static const struct type *
-lookup_atype(struct context *ctx, const struct ast_type *atype)
-{
-	const struct type *type = NULL;
-	lookup_atype_with_dimensions(ctx, &type, atype);
-	return type;
-}
-
 const struct type *
 type_store_lookup_atype(struct context *ctx, const struct ast_type *atype)
 {
 	if (atype->storage == STORAGE_NULL) {
 		return &builtin_type_null;
 	}
-	return lookup_atype(ctx, atype);
+	struct type temp = {0};
+	type_init_from_atype(ctx, &temp, atype);
+	if (temp.storage == STORAGE_ALIAS) {
+		// References to type aliases always inherit the flags that the
+		// alias was defined with
+		const struct scope_object *obj = scope_lookup(
+			ctx->scope, temp.alias.name);
+		temp.flags |= obj->type->flags;
+	}
+	return type_store_lookup_type(ctx, &temp);
 }
 
 // Compute dimensions of an incomplete type without completing it
@@ -1107,12 +1081,12 @@ type_store_lookup_with_flags(struct context *ctx,
 	}
 	struct type new = *type;
 	new.flags = flags;
-	return _type_store_lookup_type(ctx, &new, NULL);
+	return type_store_lookup_type(ctx, &new);
 }
 
 const struct type *
 type_store_lookup_pointer(struct context *ctx, struct location loc,
-	const struct type *referent, unsigned int ptrflags)
+	const struct type *referent, bool nullable)
 {
 	if (referent->storage == STORAGE_ERROR) {
 		return &builtin_type_error;
@@ -1135,7 +1109,7 @@ type_store_lookup_pointer(struct context *ctx, struct location loc,
 		.storage = STORAGE_POINTER,
 		.pointer = {
 			.referent = referent,
-			.flags = ptrflags,
+			.nullable = nullable,
 		},
 		.size = builtin_type_uintptr.size,
 		.align = builtin_type_uintptr.align,
@@ -1215,30 +1189,35 @@ type_store_lookup_slice(struct context *ctx, struct location loc,
 }
 
 const struct type *
-type_store_lookup_alias(struct context *ctx,
-		const struct type *type,
-		const struct dimensions *dims)
+type_store_lookup_alias(struct context *ctx, struct ident *ident,
+	struct ident *name, const struct type *secondary, int flags,
+	bool exported)
 {
-	return _type_store_lookup_type(ctx, type, dims);
+	struct type type = {
+		.storage = STORAGE_ALIAS,
+		.flags = flags,
+		.alias.type = secondary,
+		.alias.ident = ident,
+		.alias.name = name,
+		.alias.exported = exported,
+	};
+	return type_store_lookup_type(ctx, &type);
 }
 
 
 // Sorts members by id and deduplicates entries. Does not enforce usual tagged
 // union invariants. The returned type is not a singleton.
-static const struct type *
+static struct type *
 lookup_tagged(struct context *ctx, struct type_tagged_union *tags)
 {
-	struct type type = {
-		.storage = STORAGE_TAGGED,
-	};
+	struct type *ret = xcalloc(1, sizeof(struct type));
+	ret->storage = STORAGE_TAGGED;
 	size_t nmemb = sum_tagged_memb(ctx, tags);
 	struct type_tagged_union **tu =
 		xcalloc(nmemb, sizeof(struct type_tagged_union *));
 	size_t i = 0;
 	collect_tagged_memb(ctx, tu, tags, &i);
-	tagged_init(&type, tu, nmemb);
-	struct type *ret = xcalloc(1, sizeof(struct type));
-	*ret = type;
+	tagged_init(ret, tu, nmemb);
 	return ret;
 }
 
@@ -1246,46 +1225,14 @@ const struct type *
 type_store_lookup_tagged(struct context *ctx, struct location loc,
 		struct type_tagged_union *tags)
 {
-	const struct type *type = lookup_tagged(ctx, tags);
+	struct type *type = lookup_tagged(ctx, tags);
 	if (!enforce_tagged_invariants(ctx, loc, type)) {
 		return &builtin_type_error;
 	}
+	add_padding(&type->size, type->align);
 	return type_store_lookup_type(ctx, type);
 }
 
-const struct type *
-type_store_tagged_to_union(struct context *ctx, const struct type *tagged)
-{
-	assert(tagged->storage == STORAGE_TAGGED);
-	struct type type = {
-		.storage = STORAGE_UNION,
-		.flags = tagged->flags,
-	};
-	struct struct_field **next = &type.struct_union.fields;
-	for (const struct type_tagged_union *tu = &tagged->tagged;
-			tu; tu = tu->next) {
-		if (tu->type->size == 0) {
-			continue;
-		}
-
-		if (tu->type->size > type.size) {
-			type.size = tu->type->size;
-		}
-		if (tu->type->align > type.align) {
-			type.align = tu->type->align;
-		}
-
-		struct struct_field *sf =
-			xcalloc(1, sizeof(struct struct_field));
-		sf->name = "unnamed";
-		sf->type = tu->type;
-		sf->next = *next, *next = sf;
-		next = &sf->next;
-	}
-	type.struct_union.c_compat = true; // XXX: Unsure about this
-	return _type_store_lookup_type(ctx, &type, NULL);
-}
-
 const struct type *
 type_store_lookup_tuple(struct context *ctx, struct location loc,
 		struct type_tuple *values)
@@ -1313,15 +1260,16 @@ type_store_lookup_tuple(struct context *ctx, struct location loc,
 		if (t->type->align > type.align) {
 			type.align = t->type->align;
 		}
+		t->offset = type.size;
 		if (t->type->align != 0) {
-			t->offset = type.size;
-			if (type.size % t->type->align) {
+			if (type.size % t->type->align != 0) {
 				t->offset += t->type->align - type.size % t->type->align;
 			}
 			type.size = t->offset + t->type->size;
 		}
 	}
 	type.tuple = *values;
+	add_padding(&type.size, type.align);
 	return type_store_lookup_type(ctx, &type);
 }
 
@@ -1332,11 +1280,10 @@ type_store_lookup_enum(struct context *ctx, const struct ast_type *atype,
 	struct type type = {0};
 	type.storage = STORAGE_ENUM;
 	type.flags = atype->flags;
-	mkident(ctx, &type.alias.ident, &atype->alias, NULL);
-	identifier_dup(&type.alias.name, &atype->alias);
+	type.alias.ident = mkident(ctx, atype->alias, NULL);
+	type.alias.name = atype->alias;
 	type.alias.exported = exported;
-	type.alias.type =
-		builtin_type_for_storage(atype->_enum.storage, false);
+	type.alias.type = builtin_type_for_storage(atype->_enum.storage);
 	if (!type_is_integer(ctx, type.alias.type)
 			&& type.alias.type->storage != STORAGE_RUNE) {
 		error(ctx, atype->loc, NULL,
@@ -1382,24 +1329,6 @@ type_store_reduce_result(struct context *ctx, struct location loc,
 		bool dropped = false;
 		const struct type *it = i->type;
 
-		if (it->flags & TYPE_CONST) {
-			struct type_tagged_union **j = &in;
-			while (*j) {
-				const struct type *jt = (*j)->type;
-				if (jt == it) {
-					j = &(*j)->next;
-					continue;
-				}
-				jt = type_store_lookup_with_flags(ctx, jt,
-					jt->flags | TYPE_CONST);
-				if (jt == it) {
-					*j = (*j)->next;
-				} else {
-					j = &(*j)->next;
-				}
-			}
-		}
-
 		if (it->storage == STORAGE_NEVER || it->storage == STORAGE_ERROR) {
 			*tu = i->next;
 			continue;
@@ -1418,12 +1347,11 @@ type_store_reduce_result(struct context *ctx, struct location loc,
 			if (it->flags != jt->flags) {
 				continue;
 			}
-			if ((it->pointer.flags & PTR_NULLABLE)
-					|| (jt->pointer.flags & PTR_NULLABLE)) {
+			if (it->pointer.nullable || jt->pointer.nullable) {
 				it = type_store_lookup_pointer(ctx, loc,
-					it->pointer.referent, PTR_NULLABLE);
+					it->pointer.referent, true);
 				jt = type_store_lookup_pointer(ctx, loc,
-					jt->pointer.referent, PTR_NULLABLE);
+					jt->pointer.referent, true);
 				if (it == jt) {
 					dropped = true;
 					*tu = i->next;
@@ -1454,7 +1382,7 @@ type_store_reduce_result(struct context *ctx, struct location loc,
 	if (null != NULL && ptr != NULL) {
 		*null = (*null)->next;
 		ptr->type = type_store_lookup_pointer(ctx, loc,
-			ptr->type->pointer.referent, PTR_NULLABLE);
+			ptr->type->pointer.referent, true);
 	}
 
 	if (in == NULL) {
