@@ -10,63 +10,84 @@
 #include "util.h"
 
 const struct type *
-type_dereference(struct context *ctx, const struct type *type)
+type_dereference(struct context *ctx, const struct type *type, bool allow_nullable)
 {
 	switch (type->storage) {
 	case STORAGE_ALIAS:
 		if (type_dealias(ctx, type)->storage != STORAGE_POINTER) {
 			return type;
 		}
-		return type_dereference(ctx, type_dealias(ctx, type));
+		return type_dereference(ctx, type_dealias(ctx, type), allow_nullable);
 	case STORAGE_POINTER:
-		if (type->pointer.flags & PTR_NULLABLE) {
+		if (!allow_nullable && type->pointer.nullable) {
 			return NULL;
 		}
-		return type_dereference(ctx, type->pointer.referent);
+		return type_dereference(ctx, type->pointer.referent, allow_nullable);
 	default:
 		return type;
 	}
 }
 
-void
+static const struct scope_object *
 complete_alias(struct context *ctx, struct type *type)
 {
 	assert(type->storage == STORAGE_ALIAS);
 	const struct scope_object *obj =
-		scope_lookup(ctx->scope, &type->alias.name);
+		scope_lookup(ctx->scope, type->alias.name);
 	assert(obj != NULL);
 	assert(obj->otype == O_TYPE || obj->otype == O_SCAN);
-	struct incomplete_declaration *idecl =
-		(struct incomplete_declaration *)obj;
-	assert(idecl->type == IDECL_DECL);
+	assert(obj->idecl->type == IDECL_DECL);
 
-	if (idecl->dealias_in_progress) {
-		char *identstr = identifier_unparse(&idecl->obj.name);
-		error(ctx, idecl->decl.loc, NULL,
-			"Circular dependency for '%s'", identstr);
-		free(identstr);
-		type->alias.type = &builtin_type_error;
-		return;
+	if (!obj->idecl->dealias_in_progress) {
+		obj->idecl->dealias_in_progress = true;
+		type->alias.type = type_store_lookup_atype(
+			ctx, obj->idecl->decl.type.type);
+		obj->idecl->dealias_in_progress = false;
 	}
-	idecl->dealias_in_progress = true;
-	type->alias.type = type_store_lookup_atype(ctx, idecl->decl.type.type);
-	idecl->dealias_in_progress = false;
+	return obj;
 }
 
 const struct type *
-type_dealias(struct context *ctx, const struct type *type)
+type_dealias(struct context *ctx, const struct type *_type)
 {
+	struct type *type = (struct type *)_type;
 	while (type->storage == STORAGE_ALIAS) {
 		if (type->alias.type == NULL) {
 			// gen et al. don't have access to the check context,
 			// but by that point all aliases should already be fully
 			// scanned
 			assert(ctx != NULL);
+			const struct scope_object *obj =
+				complete_alias(ctx, type);
+			if (type->alias.type == NULL) {
+				char *identstr = ident_unparse(obj->name);
+				error(ctx, obj->idecl->decl.loc, NULL,
+					"Circular dependency for '%s'",
+					identstr);
+				free(identstr);
+				type->alias.type = &builtin_type_error;
+			}
+		}
+		type = (struct type *)type->alias.type;
+	}
+	return type;
+}
+
+// checks if a type is `done`, or an alias thereof, without erroring out when a
+// "circular dependency" is encountered (since that means the type isn't `done`)
+bool
+type_is_done(struct context *ctx, const struct type *type)
+{
+	while (type->storage == STORAGE_ALIAS) {
+		if (type->alias.type == NULL) {
 			complete_alias(ctx, (struct type *)type);
+			if (type->alias.type == NULL) {
+				return false;
+			}
 		}
 		type = type->alias.type;
 	}
-	return type;
+	return type->storage == STORAGE_DONE;
 }
 
 const struct struct_field *
@@ -149,7 +170,7 @@ type_storage_unparse(enum type_storage storage)
 	case STORAGE_ERROR:
 		return "invalid";
 	case STORAGE_FCONST:
-		return "fconst";
+		return "flexible float";
 	case STORAGE_FUNCTION:
 		return "function";
 	case STORAGE_I16:
@@ -161,11 +182,13 @@ type_storage_unparse(enum type_storage storage)
 	case STORAGE_I8:
 		return "i8";
 	case STORAGE_ICONST:
-		return "iconst";
+		return "flexible integer";
 	case STORAGE_INT:
 		return "int";
 	case STORAGE_NEVER:
 		return "never";
+	case STORAGE_NOMEM:
+		return "nomem";
 	case STORAGE_NULL:
 		return "null";
 	case STORAGE_OPAQUE:
@@ -173,7 +196,7 @@ type_storage_unparse(enum type_storage storage)
 	case STORAGE_POINTER:
 		return "pointer";
 	case STORAGE_RCONST:
-		return "rconst";
+		return "flexible rune";
 	case STORAGE_RUNE:
 		return "rune";
 	case STORAGE_SIZE:
@@ -221,6 +244,7 @@ type_is_integer(struct context *ctx, const struct type *type)
 	case STORAGE_ARRAY:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_POINTER:
 	case STORAGE_SLICE:
@@ -269,6 +293,7 @@ type_is_numeric(struct context *ctx, const struct type *type)
 	case STORAGE_ARRAY:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_POINTER:
 	case STORAGE_SLICE:
@@ -332,6 +357,7 @@ type_is_signed(struct context *ctx, const struct type *type)
 	case STORAGE_ERROR: // XXX?
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_POINTER:
 	case STORAGE_SLICE:
@@ -395,6 +421,7 @@ type_hash(const struct type *type)
 	case STORAGE_I64:
 	case STORAGE_INT:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_NULL:
 	case STORAGE_OPAQUE:
 	case STORAGE_RUNE:
@@ -414,11 +441,7 @@ type_hash(const struct type *type)
 		hash = fnv1a(hash, type->alias.type->storage);
 		/* fallthrough */
 	case STORAGE_ALIAS:
-		for (const struct identifier *ident = &type->alias.ident; ident;
-				ident = ident->ns) {
-			hash = fnv1a_s(hash, ident->name);
-			hash = fnv1a(hash, 0);
-		}
+		hash = ident_hash(hash, type->alias.ident);
 		break;
 	case STORAGE_ARRAY:
 		hash = fnv1a_u32(hash, type_hash(type->array.members));
@@ -443,7 +466,7 @@ type_hash(const struct type *type)
 		hash = fnv1a(hash, type->flexible.id);
 		break;
 	case STORAGE_POINTER:
-		hash = fnv1a(hash, type->pointer.flags);
+		hash = fnv1a(hash, (unsigned char) type->pointer.nullable);
 		hash = fnv1a_u32(hash, type_hash(type->pointer.referent));
 		break;
 	case STORAGE_SLICE:
@@ -791,12 +814,10 @@ struct_subtype(struct context *ctx,
 	if (from->storage != STORAGE_STRUCT) {
 		return false;
 	}
-	for (struct struct_field *f = from->struct_union.fields; f;
-			f = f->next) {
-		if (f->offset == 0) {
-			return f->type == to
-				|| struct_subtype(ctx, to, type_dealias(ctx, f->type));
-		}
+	for (struct struct_field *f = from->struct_union.fields;
+			f && f->offset == 0; f = f->next) {
+		return f->type == to
+			|| struct_subtype(ctx, to, type_dealias(ctx, f->type));
 	}
 	return false;
 }
@@ -814,8 +835,7 @@ type_is_assignable(struct context *ctx,
 	// const and non-const types are mutually assignable
 	struct type _to, _from;
 	to = strip_flags(to, &_to), from = strip_flags(from, &_from);
-	if (to->id == from->id && to->storage != STORAGE_VOID
-			&& to->storage != STORAGE_DONE) {
+	if (to->id == from->id && to->storage != STORAGE_VOID) {
 		return true;
 	}
 
@@ -858,7 +878,7 @@ type_is_assignable(struct context *ctx,
 		to_secondary = strip_flags(to_secondary, &_to_secondary);
 		switch (from->storage) {
 		case STORAGE_NULL:
-			return to->pointer.flags & PTR_NULLABLE;
+			return to->pointer.nullable;
 		case STORAGE_POINTER:
 			from_secondary = from->pointer.referent;
 			from_secondary = strip_flags(from_secondary, &_from_secondary);
@@ -879,8 +899,8 @@ type_is_assignable(struct context *ctx,
 				}
 				break;
 			}
-			if (from->pointer.flags & PTR_NULLABLE) {
-				return to->pointer.flags & PTR_NULLABLE;
+			if (from->pointer.nullable) {
+				return to->pointer.nullable;
 			}
 			return true;
 		default:
@@ -891,7 +911,6 @@ type_is_assignable(struct context *ctx,
 		assert(to->alias.type);
 		return type_is_assignable(ctx, to->alias.type, from);
 	case STORAGE_VOID:
-	case STORAGE_DONE:
 		return to_orig->id == from_orig->id &&
 			(from_orig->flags & TYPE_ERROR)	== (to_orig->flags & TYPE_ERROR);
 	case STORAGE_SLICE:
@@ -934,10 +953,12 @@ type_is_assignable(struct context *ctx,
 	// The following types are only assignable from themselves, and are
 	// handled above:
 	case STORAGE_BOOL:
+	case STORAGE_DONE:
 	case STORAGE_ENUM:
 	case STORAGE_F32:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_NULL:
 	case STORAGE_OPAQUE:
 	case STORAGE_RUNE:
@@ -983,7 +1004,7 @@ is_castable_with_tagged(struct context *ctx,
 const struct type *
 type_is_castable(struct context *ctx, const struct type *to, const struct type *from)
 {
-	if (to->storage == STORAGE_VOID || to->storage == STORAGE_DONE) {
+	if (to->storage == STORAGE_VOID) {
 		if (type_is_flexible(from)) {
 			lower_flexible(ctx, from, NULL);
 		}
@@ -1055,7 +1076,10 @@ type_is_castable(struct context *ctx, const struct type *to, const struct type *
 		return type_is_integer(ctx, to)
 			? to_orig : NULL;
 	case STORAGE_ENUM:
-		return to->storage == STORAGE_ENUM || type_is_integer(ctx, from)
+		if (from->alias.type->storage == STORAGE_RUNE) {
+			return to->storage == STORAGE_RUNE ? to_orig : NULL;
+		}
+		return to->storage == STORAGE_ENUM || type_is_integer(ctx, to)
 			? to_orig : NULL;
 	case STORAGE_F32:
 	case STORAGE_F64:
@@ -1090,13 +1114,14 @@ type_is_castable(struct context *ctx, const struct type *to, const struct type *
 	case STORAGE_BOOL:
 	case STORAGE_VOID:
 	case STORAGE_DONE:
+	case STORAGE_NEVER:
+	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_FUNCTION:
 	case STORAGE_TUPLE:
 	case STORAGE_STRUCT:
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
-	case STORAGE_NEVER:
 		return NULL;
 	case STORAGE_ERROR:
 	case STORAGE_TAGGED:
@@ -1111,82 +1136,61 @@ void
 builtin_types_init(const char *target)
 {
 	if (strcmp(target, "aarch64") == 0) {
+		builtin_type_f64.align = 8;
 		builtin_type_int.size = 4;
 		builtin_type_int.align = 4;
 		builtin_type_uint.size = 4;
 		builtin_type_uint.align = 4;
 		builtin_type_uintptr.size = 8;
 		builtin_type_uintptr.align = 8;
+		builtin_type_i64.align = 8;
+		builtin_type_u64.align = 8;
 		builtin_type_null.size = 8;
 		builtin_type_null.align = 8;
 		builtin_type_size.size = 8;
 		builtin_type_size.align = 8;
-		builtin_type_const_int.size = 4;
-		builtin_type_const_int.align = 4;
-		builtin_type_const_uint.size = 4;
-		builtin_type_const_uint.align = 4;
-		builtin_type_const_uintptr.size = 8;
-		builtin_type_const_uintptr.align = 8;
-		builtin_type_const_size.size = 8;
-		builtin_type_const_size.align = 8;
 		builtin_type_str.size = 24;
 		builtin_type_str.align = 8;
-		builtin_type_const_str.size = 24;
-		builtin_type_const_str.align = 8;
 		builtin_type_valist.size = 32;
 		builtin_type_valist.align = 8;
 	} else if (strcmp(target, "riscv64") == 0) {
+		builtin_type_f64.align = 8;
 		builtin_type_int.size = 4;
 		builtin_type_int.align = 4;
 		builtin_type_uint.size = 4;
 		builtin_type_uint.align = 4;
 		builtin_type_uintptr.size = 8;
 		builtin_type_uintptr.align = 8;
+		builtin_type_i64.align = 8;
+		builtin_type_u64.align = 8;
 		builtin_type_null.size = 8;
 		builtin_type_null.align = 8;
 		builtin_type_size.size = 8;
 		builtin_type_size.align = 8;
-		builtin_type_const_int.size = 4;
-		builtin_type_const_int.align = 4;
-		builtin_type_const_uint.size = 4;
-		builtin_type_const_uint.align = 4;
-		builtin_type_const_uintptr.size = 8;
-		builtin_type_const_uintptr.align = 8;
-		builtin_type_const_size.size = 8;
-		builtin_type_const_size.align = 8;
 		builtin_type_str.size = 24;
 		builtin_type_str.align = 8;
-		builtin_type_const_str.size = 24;
-		builtin_type_const_str.align = 8;
 		builtin_type_valist.size = 8;
 		builtin_type_valist.align = 8;
 	} else if (strcmp(target, "x86_64") == 0) {
+		builtin_type_f64.align = 8;
 		builtin_type_int.size = 4;
 		builtin_type_int.align = 4;
 		builtin_type_uint.size = 4;
 		builtin_type_uint.align = 4;
 		builtin_type_uintptr.size = 8;
 		builtin_type_uintptr.align = 8;
+		builtin_type_i64.align = 8;
+		builtin_type_u64.align = 8;
 		builtin_type_null.size = 8;
 		builtin_type_null.align = 8;
 		builtin_type_size.size = 8;
 		builtin_type_size.align = 8;
-		builtin_type_const_int.size = 4;
-		builtin_type_const_int.align = 4;
-		builtin_type_const_uint.size = 4;
-		builtin_type_const_uint.align = 4;
-		builtin_type_const_uintptr.size = 8;
-		builtin_type_const_uintptr.align = 8;
-		builtin_type_const_size.size = 8;
-		builtin_type_const_size.align = 8;
 		builtin_type_str.size = 24;
 		builtin_type_str.align = 8;
-		builtin_type_const_str.size = 24;
-		builtin_type_const_str.align = 8;
 		builtin_type_valist.size = 24;
 		builtin_type_valist.align = 8;
 	} else {
-		xfprintf(stderr, "Unsupported or unrecognized target: %s", target);
+		xfprintf(stderr, "Unsupported or unrecognized target: %s\n", target);
 		exit(EXIT_USER);
 	}
 	struct type *builtins[] = {
@@ -1196,18 +1200,8 @@ builtin_types_init(const char *target)
 		&builtin_type_u8, &builtin_type_u16, &builtin_type_u32,
 		&builtin_type_u64, &builtin_type_uint, &builtin_type_uintptr,
 		&builtin_type_null, &builtin_type_rune, &builtin_type_size,
-		&builtin_type_void, &builtin_type_done,
-		&builtin_type_const_bool, &builtin_type_const_f32,
-		&builtin_type_const_f64, &builtin_type_const_i8,
-		&builtin_type_const_i16, &builtin_type_const_i32,
-		&builtin_type_const_i64, &builtin_type_const_int,
-		&builtin_type_const_u8, &builtin_type_const_u16,
-		&builtin_type_const_u32, &builtin_type_const_u64,
-		&builtin_type_const_uint, &builtin_type_const_uintptr,
-		&builtin_type_const_rune, &builtin_type_const_size,
-		&builtin_type_const_void, &builtin_type_const_done,
-		&builtin_type_str, &builtin_type_const_str,
-		&builtin_type_valist,
+		&builtin_type_void, &builtin_type_done, &builtin_type_nomem,
+		&builtin_type_str, &builtin_type_valist,
 	};
 	for (size_t i = 0; i < sizeof(builtins) / sizeof(builtins[0]); ++i) {
 		builtins[i]->id = type_hash(builtins[i]);
@@ -1233,7 +1227,6 @@ builtin_type_f32 = {
 builtin_type_f64 = {
 	.storage = STORAGE_F64,
 	.size = 8,
-	.align = 8,
 },
 builtin_type_i8 = {
 	.storage = STORAGE_I8,
@@ -1253,7 +1246,6 @@ builtin_type_i32 = {
 builtin_type_i64 = {
 	.storage = STORAGE_I64,
 	.size = 8,
-	.align = 8,
 },
 builtin_type_int = {
 	.storage = STORAGE_INT,
@@ -1263,6 +1255,12 @@ builtin_type_never = {
 	.size = SIZE_UNDEFINED,
 	.align = ALIGN_UNDEFINED,
 },
+builtin_type_nomem = {
+	.storage = STORAGE_NOMEM,
+	.flags = TYPE_ERROR,
+	.size = 0,
+	.align = 0,
+},
 builtin_type_opaque = {
 	.storage = STORAGE_OPAQUE,
 	.size = SIZE_UNDEFINED,
@@ -1286,7 +1284,6 @@ builtin_type_u32 = {
 builtin_type_u64 = {
 	.storage = STORAGE_U64,
 	.size = 8,
-	.align = 8,
 },
 builtin_type_uint = {
 	.storage = STORAGE_UINT,
@@ -1315,127 +1312,9 @@ builtin_type_done = {
 	.size = 0,
 	.align = 0,
 },
-builtin_type_const_bool = {
-	.storage = STORAGE_BOOL,
-	.flags = TYPE_CONST,
-	.size = 1,
-	.align = 1,
-},
-builtin_type_const_f32 = {
-	.storage = STORAGE_F32,
-	.flags = TYPE_CONST,
-	.size = 4,
-	.align = 4,
-},
-builtin_type_const_f64 = {
-	.storage = STORAGE_F64,
-	.flags = TYPE_CONST,
-	.size = 8,
-	.align = 8,
-},
-builtin_type_const_i8 = {
-	.storage = STORAGE_I8,
-	.flags = TYPE_CONST,
-	.size = 1,
-	.align = 1,
-},
-builtin_type_const_i16 = {
-	.storage = STORAGE_I16,
-	.flags = TYPE_CONST,
-	.size = 2,
-	.align = 2,
-},
-builtin_type_const_i32 = {
-	.storage = STORAGE_I32,
-	.flags = TYPE_CONST,
-	.size = 4,
-	.align = 4,
-},
-builtin_type_const_i64 = {
-	.storage = STORAGE_I64,
-	.flags = TYPE_CONST,
-	.size = 8,
-	.align = 8,
-},
-builtin_type_const_int = {
-	.storage = STORAGE_INT,
-	.flags = TYPE_CONST,
-},
-builtin_type_const_never = {
-	.storage = STORAGE_NEVER,
-	.flags = TYPE_CONST,
-	.size = SIZE_UNDEFINED,
-	.align = ALIGN_UNDEFINED,
-},
-builtin_type_const_opaque = {
-	.storage = STORAGE_OPAQUE,
-	.flags = TYPE_CONST,
-	.size = SIZE_UNDEFINED,
-	.align = ALIGN_UNDEFINED,
-},
-builtin_type_const_u8 = {
-	.storage = STORAGE_U8,
-	.flags = TYPE_CONST,
-	.size = 1,
-	.align = 1,
-},
-builtin_type_const_u16 = {
-	.storage = STORAGE_U16,
-	.flags = TYPE_CONST,
-	.size = 2,
-	.align = 2,
-},
-builtin_type_const_u32 = {
-	.storage = STORAGE_U32,
-	.flags = TYPE_CONST,
-	.size = 4,
-	.align = 4,
-},
-builtin_type_const_u64 = {
-	.storage = STORAGE_U64,
-	.flags = TYPE_CONST,
-	.size = 8,
-	.align = 8,
-},
-builtin_type_const_uint = {
-	.storage = STORAGE_UINT,
-	.flags = TYPE_CONST,
-},
-builtin_type_const_uintptr = {
-	.storage = STORAGE_UINTPTR,
-	.flags = TYPE_CONST,
-},
-builtin_type_const_rune = {
-	.storage = STORAGE_RUNE,
-	.flags = TYPE_CONST,
-	.size = 4,
-	.align = 4,
-},
-builtin_type_const_size = {
-	.storage = STORAGE_SIZE,
-	.flags = TYPE_CONST,
-},
-builtin_type_const_void = {
-	.storage = STORAGE_VOID,
-	.flags = TYPE_CONST,
-	.size = 0,
-	.align = 0,
-},
-builtin_type_const_done = {
-	.storage = STORAGE_DONE,
-	.flags = TYPE_CONST,
-	.size = 0,
-	.align = 0,
-};
-
-// Others
-struct type builtin_type_str = {
+builtin_type_str = {
 	.storage = STORAGE_STRING,
 },
-builtin_type_const_str = {
-	.storage = STORAGE_STRING,
-	.flags = TYPE_CONST,
-},
 builtin_type_valist = {
 	.storage = STORAGE_VALIST,
 };
